#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.24 23:00:00                  #
# ================================================== #

from PySide6.QtGui import Qt, QIcon
from PySide6.QtWidgets import QVBoxLayout, QHBoxLayout, QWidget, QSizePolicy

from pygpt_net.ui.widget.element.labels import HelpLabel, TitleLabel
from pygpt_net.ui.widget.option.prompt import PromptTextarea
from pygpt_net.ui.widget.option.toggle_label import ToggleLabel
from pygpt_net.utils import trans

class Prompt:
    def __init__(self, window=None):
        """
        Toolbox UI

        :param window: Window instance
        """
        self.window = window

    def setup(self) -> QWidget:
        """
        Setup system prompt

        :return: QWidget
        """
        w = self.window
        nodes = w.ui.nodes

        nodes['toolbox.prompt.label'] = TitleLabel(trans("toolbox.prompt"))

        nodes['cmd.enabled'] = ToggleLabel(
            trans('cmd.enabled'),
            label_position="left",
            icon=":/icons/build.svg",
            parent=w
        )
        box = nodes['cmd.enabled'].box
        box.toggled.connect(w.controller.chat.common.toggle_cmd)
        box.setToolTip(trans('cmd.tip'))
        box.setIcon(QIcon(":/icons/add.svg"))

        header_widget = QWidget()
        header_layout = QHBoxLayout(header_widget)
        header_layout.addWidget(nodes['toolbox.prompt.label'])
        header_layout.addStretch(1)
        header_layout.addWidget(nodes['cmd.enabled'])
        header_layout.setContentsMargins(5, 0, 10, 0)

        option = w.controller.presets.editor.get_option('prompt')
        nodes['preset.prompt'] = PromptTextarea(w, 'preset', 'prompt', option)
        nodes['preset.prompt'].setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        nodes['tip.toolbox.prompt'] = HelpLabel(trans('tip.toolbox.prompt'), w)
        nodes['tip.toolbox.prompt'].setAlignment(Qt.AlignCenter)

        layout_widget = QWidget()
        layout = QVBoxLayout(layout_widget)
        layout.addWidget(header_widget)
        layout.addWidget(nodes['preset.prompt'])
        layout.addWidget(nodes['tip.toolbox.prompt'])
        layout.setContentsMargins(2, 5, 5, 5)

        layout_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        return layout_widget


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.15 01:00:00                  #
# ================================================== #

from typing import Optional, Dict, Any, Union

from llama_index.embeddings.huggingface_api import (
    HuggingFaceInferenceAPIEmbedding as _HFEmbed,
)

class HuggingFaceInferenceAPIEmbeddingWithProxy(_HFEmbed):
    def __init__(
        self,
        *,
        proxy: Optional[Union[str, Dict[str, str]]] = None,
        trust_env: Optional[bool] = None,
        **kwargs: Any,
    ) -> None:
        """
        proxy: e.g. "http://user:pass@host:3128" albo {"http": "...", "https": "..."}
        trust_env: AsyncInferenceClient; if True, then use the system's proxy settings
        """

        if "api_key" in kwargs and "token" not in kwargs:
            kwargs["token"] = kwargs.pop("api_key")
        if "model" in kwargs and "model_name" not in kwargs:
            kwargs["model_name"] = kwargs.pop("model")

        super().__init__(**kwargs)

        # if no proxy and no trust_env, then do nothing
        if proxy is None and trust_env is None:
            return

        # dict proxies for huggingface_hub.*
        hf_proxies: Optional[Union[str, Dict[str, str]]] = None
        if proxy is not None:
            hf_proxies = {"http": proxy, "https": proxy} if isinstance(proxy, str) else proxy

        base_kwargs: Dict[str, Any] = {}

        for src, dst in [
            ("model_name", "model"),
            ("token", "token"),
            ("timeout", "timeout"),
            ("headers", "headers"),
            ("cookies", "cookies"),
            ("base_url", "base_url"),
        ]:
            if hasattr(self, src):
                val = getattr(self, src)
                if val is not None and val != "":
                    base_kwargs[dst] = val

        from huggingface_hub import InferenceClient, AsyncInferenceClient

        sync_kwargs = dict(base_kwargs)
        async_kwargs = dict(base_kwargs)

        if hf_proxies is not None:
            sync_kwargs["proxies"] = hf_proxies
            async_kwargs["proxies"] = hf_proxies
        if trust_env is not None:
            async_kwargs["trust_env"] = trust_env  # default False w AsyncInferenceClient

        sync_client = InferenceClient(**sync_kwargs)
        async_client = AsyncInferenceClient(**async_kwargs)

        for name, client in (
            ("_sync_client", sync_client),
            ("_client", sync_client),
        ):
            if hasattr(self, name):
                setattr(self, name, client)

        for name, client in (
            ("_async_client", async_client),
            ("_aclient", async_client),
        ):
            if hasattr(self, name):
                setattr(self, name, client)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.27 07:00:00                  #
# ================================================== #

from pygpt_net.utils import trans


class UI:
    def __init__(self, window=None):
        """
        Audio/voice UI controller

        :param window: Window instance
        """
        self.window = window
        self.recording = False

        self._input_bar = None
        self._input_btn = None
        self._input_control_bar = None
        self._input_control_btn = None
        self._output_bar = None
        self._output_btn = None

    def get_input_bar(self):
        """
        Get input bar widget

        :return: input bar widget
        """
        if self._input_bar is not None:
            return self._input_bar
        if "audio.input.btn" not in self.window.ui.plugin_addon:
            return None
        self._input_bar = self.window.ui.plugin_addon['audio.input.btn'].bar
        return self._input_bar

    def get_input_control_bar(self):
        """
        Get input control bar widget

        :return: input control bar widget
        """
        if self._input_control_bar is not None:
            return self._input_control_bar
        if 'voice.control.btn' not in self.window.ui.nodes:
            return None
        self._input_control_bar = self.window.ui.nodes['voice.control.btn'].bar
        return self._input_control_bar

    def get_output_bar(self):
        """
        Get output bar widget

        :return: output bar widget
        """
        if self._output_bar is not None:
            return self._output_bar
        if "audio.output.bar" not in self.window.ui.plugin_addon:
            return None
        self._output_bar = self.window.ui.plugin_addon['audio.output.bar']
        return self._output_bar

    def get_input_btn(self):
        """
        Get input bar widget

        :return: input btn widget
        """
        if self._input_btn is not None:
            return self._input_btn
        if "audio.input.btn" not in self.window.ui.plugin_addon:
            return None
        self._input_btn = self.window.ui.plugin_addon['audio.input.btn']
        return self._input_btn

    def get_input_control_btn(self):
        """
        Get input control btn widget

        :return: input control btn widget
        """
        if self._input_control_btn is not None:
            return self._input_control_btn
        if 'voice.control.btn' not in self.window.ui.nodes:
            return None
        self._input_control_btn = self.window.ui.nodes['voice.control.btn']
        return self._input_control_btn

    def get_output_btn(self):
        """
        Get output btn widget

        :return: output btn widget
        """
        if self._output_btn is not None:
            return self._output_btn
        if "audio.output" not in self.window.ui.plugin_addon:
            return None
        self._output_btn = self.window.ui.plugin_addon['audio.output']
        return self._output_btn

    # --- Input events ---

    def on_input_volume_change(self, value: int, mode: str = 'input'):
        """
        Input volume slider change event

        :param value: slider value
        :param mode: 'input' or 'control' mode
        """
        bar = self.get_output_bar()
        if bar:
            bar.setLevel(value)

    def on_input_device_change(self, index: int):
        """
        Input device combo change event

        :param index: combo index
        """
        pass

    def on_input_enable(self, mode: str = 'input'):
        """
        Input enable checkbox change event

        :param mode: 'input' or 'control' mode
        """
        self.window.ui.nodes['input'].set_icon_visible("mic", True)
        if mode == "input":
            return
        btn = self.get_input_btn() if mode == 'input' else self.get_input_control_btn()
        if btn:
            btn.setVisible(True)

    def on_input_disable(self, mode: str = 'input'):
        """
        Input disabled button click event

        :param mode: 'input' or 'control' mode
        """
        self.window.ui.nodes['input'].set_icon_visible("mic", False)
        if mode == "input":
            return
        btn = self.get_input_btn() if mode == 'input' else self.get_input_control_btn()
        if btn:
            btn.setVisible(False)

    def on_input_continuous_enable(self, checked: bool, mode: str = 'input'):
        """
        Input enable checkbox change event

        :param checked: checkbox state
        :param mode: 'input' or 'control' mode
        """
        btn = self.get_input_btn() if mode == 'input' else self.get_input_control_btn()
        if btn:
            btn.notepad_footer.setVisible(True)

    def on_input_continuous_disable(self, mode: str = 'input'):
        """
        Input disabled button click event

        :param mode: 'input' or 'control' mode
        """
        btn = self.get_input_btn() if mode == 'input' else self.get_input_control_btn()
        if btn:
            btn.notepad_footer.setVisible(False)

    def on_input_begin(self, mode: str = 'input'):
        """
        Input begin button click event

        :param mode: 'input' or 'control' mode
        """
        self.recording = True
        self.window.ui.nodes['input'].set_icon_state("mic", True)
        if mode in ["input", "realtime"]:
            self.window.controller.chat.common.lock_input()
            return
        btn = self.get_input_btn() if mode == 'input' else self.get_input_control_btn()
        btn.btn_toggle.setText(trans('audio.speak.btn.stop'))
        btn.btn_toggle.setToolTip(trans('audio.speak.btn.stop.tooltip'))

    def on_input_end(self, mode: str = 'input'):
        """
        Input end button click event

        :param mode: 'input' or 'control' mode
        """
        self.recording = False
        self.window.ui.nodes['input'].set_icon_state("mic", False)
        if mode in ["input", "realtime"]:
            self.window.controller.chat.common.unlock_input()
            return
        btn = self.get_input_btn() if mode == 'input' else self.get_input_control_btn()
        btn.btn_toggle.setText(trans('audio.speak.btn'))
        btn.btn_toggle.setToolTip(trans('audio.speak.btn.tooltip'))

    def on_input_cancel(self):
        """
        Input cancel button click event
        """
        pass

    # --- Output events ---

    def on_output_volume_change(self, value: int):
        """
        Output volume slider change event
        :param value: slider value
        """
        if self.recording:
            return
        bar = self.get_output_bar()
        if bar:
            bar.setLevel(value)

    def on_output_device_change(self, index: int):
        """
        Output device combo change event

        :param index: combo index
        """
        pass

    def on_output_enable(self):
        """
        Output enable checkbox change event
        """
        pass

    def on_output_disable(self):
        """
        Output disabled button click event
        """
        pass

    def on_output_begin(self):
        """
        Output begin button click event
        """
        self.window.controller.audio.start_speaking()

    def on_output_end(self):
        """
        Output end button click event
        """
        pass

    def on_output_cancel(self):
        """
        Output cancel button click event
        """
        pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.09 15:00:00                  #
# ================================================== #
import os

from pygpt_net.utils import trans


class Text:
    def __init__(self, window=None):
        """
        Text helpers

        :param window: Window instance
        """
        self.window = window
        self.lang_list = None  # cache for language choices

    def get_language_choices(self) -> list:
        """
        Get available language choices

        :return: list of dictionaries with language codes and names
        """
        if self.lang_list is not None:
            return self.lang_list  # return cached choices

        choices = []
        choices.append({"-": trans("translator.search.auto")})
        csv_path = os.path.join(self.window.core.config.get_app_path(), 'data', 'languages.csv')
        if os.path.exists(csv_path):
            with open(csv_path, 'r', encoding='utf-8') as file:
                for line in file.readlines()[1:]:
                    parts = line.strip().split(',')
                    if len(parts) >= 4:
                        lang_code = parts[0].strip()
                        lang_name = parts[3].strip()
                        lang_orig_name = parts[4].strip()
                        name = f"{lang_name} ({lang_orig_name})" if lang_orig_name else lang_name
                        name = name.replace("'", "").replace('"', "")
                        choices.append({lang_code: name})

        # sort choices by language name
        choices.sort(key=lambda x: list(x.values())[0].lower())

        self.lang_list = choices  # cache the choices for later use
        return choices

    def get_language_name(self, lang_code: str) -> str:
        """
        Get language name by code

        :param lang_code: language code
        :return: language name or empty string if not found
        """
        choices = self.get_language_choices()
        for choice in choices:
            if lang_code in choice:
                return choice[lang_code]
        return ""

    def find_lang_id_by_search_string(self, search_string: str) -> str:
        """
        Find language code by search string

        :param search_string: search string
        :return: language code or empty string if not found
        """
        choices = self.get_language_choices()
        to_search = search_string.strip().lower()
        for choice in choices:
            lang_code = list(choice.keys())[0]
            lang_name = choice[lang_code].strip().lower()
            if lang_name.startswith(to_search):
                return lang_code
        for choice in choices:
            lang_code = list(choice.keys())[0]
            lang_name = choice[lang_code].strip().lower()
            if to_search in lang_name:
                return lang_code
        return ""


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.07 05:00:00                  #
# ================================================== #

import io
from dataclasses import dataclass, field
from typing import Any, Optional

from pygpt_net.utils import trans


@dataclass(slots=True)
class PidData:
    """Pid Data"""
    # Required/primary data
    pid: Any
    meta: Optional[Any] = None

    # Collections
    images_appended: list = field(default_factory=list)
    urls_appended: list = field(default_factory=list)
    files_appended: list = field(default_factory=list)

    # Internal buffers (excluded from repr to avoid large dumps)
    _buffer: io.StringIO = field(default_factory=io.StringIO, repr=False)
    _live_buffer: io.StringIO = field(default_factory=io.StringIO, repr=False)
    _html: io.StringIO = field(default_factory=io.StringIO, repr=False)
    _document: io.StringIO = field(default_factory=io.StringIO, repr=False)
    _fence_tail: str = field(default="", repr=False)
    _fence_open: bool = field(default=False, repr=False)
    _fence_char: str = field(default="", repr=False)

    # Flags/state
    is_cmd: bool = False
    initialized: bool = False
    loaded: bool = False
    item: Optional[Any] = None
    use_buffer: bool = False

    # Names
    name_user: str = field(default_factory=lambda: trans("chat.name.user"))
    name_bot: str = field(default_factory=lambda: trans("chat.name.bot"))

    # Throttling / timing
    last_time_called: float = 0.0
    cooldown: float = 1 / 6
    throttling_min_chars: int = 5000

    # Misc
    header: Optional[Any] = None

    @property
    def buffer(self) -> str:
        return self._buffer.getvalue()

    @buffer.setter
    def buffer(self, value: str):
        if value is None or value == "":
            self._buffer.close()
            self._buffer = io.StringIO()
        self._buffer.seek(0)
        self._buffer.truncate(0)
        if value:
            self._buffer.write(value)

    def append_buffer(self, text: str):
        self._buffer.write(text)

    @property
    def live_buffer(self) -> str:
        return self._live_buffer.getvalue()

    @live_buffer.setter
    def live_buffer(self, value: str):
        if value is None or value == "":
            self._live_buffer.close()
            self._live_buffer = io.StringIO()
        self._live_buffer.seek(0)
        self._live_buffer.truncate(0)
        if value:
            self._live_buffer.write(value)

    def append_live_buffer(self, text: str):
        self._live_buffer.write(text)

    @property
    def html(self) -> str:
        return self._html.getvalue()

    @html.setter
    def html(self, value: str):
        if value is None or value == "":
            self._html.close()
            self._html = io.StringIO()
        self._html.seek(0)
        self._html.truncate(0)
        if value:
            self._html.write(value)

    def append_html(self, text: str):
        self._html.write(text)

    @property
    def document(self) -> str:
        return self._document.getvalue()

    @document.setter
    def document(self, value: str):
        if value is None or value == "":
            self._document.close()
            self._document = io.StringIO()
        self._document.seek(0)
        self._document.truncate(0)
        if value:
            self._document.write(value)

    def append_document(self, text: str):
        self._document.write(text)

    def clear(self, all: bool = False):
        """
        Clear buffers and other data

        :param all: If True, clear all data, otherwise only buffers
        """
        for name in ("_buffer", "_live_buffer", "_html", "_document"):
            try:
                getattr(self, name).close()
            except Exception:
                pass
            setattr(self, name, io.StringIO())

        if all:
            self.item = None
            self.images_appended.clear()
            self.urls_appended.clear()
            self.files_appended.clear()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.12 20:00:00                  #
# ================================================== #

import asyncio
import random
import string
import time
from dataclasses import dataclass
from typing import AsyncIterator, Iterator, Literal, Optional, Union, Dict, Any

APITypes = Literal["raw", "chat", "responses", "respones"]  # with typo alias
ChunkTypes = Literal["str", "code"]

@dataclass
class StreamConfig:
    api: APITypes = "raw"
    chunk: ChunkTypes = "str"
    cps: int = 16                 # chunks per second
    min_newline: int = 30         # minimum distance to next \n
    include_chat_role: bool = True  # in chat: first delta with {"role": "assistant"}

class FakeOpenAIStream:
    """
    Fake chunk streamer:
      - api: "raw" | "chat" | "responses" (alias: "respones")
      - chunk: "str" | "code"
      - stop() stops the stream
    Returned elements:
      - raw:        str (text only)
      - chat:       dict in chat.completion.chunk style (choices[0].delta.*)
      - responses:  dict of response.output_text.delta type (with delta + indices)
    """
    def __init__(self, *, rng_seed: Optional[int] = None, model_name: str = "gpt-4o-mini-fake", code_path: Optional[str] = None):
        self._rng = random.Random(rng_seed)
        self._stopped = False
        self._model = model_name
        self._run_id = ""
        self._msg_id = ""  # for Responses API
        # State for injecting new lines
        self._chars_since_nl = 0
        self._next_nl_at = 80
        self._code_path = code_path
        self._code_src = None
        self._code_pos = 0
        self._code_mode = False
        if code_path:
            try:
                with open(code_path, "r", encoding="utf-8") as f:
                    self._code_src = f.read()
            except Exception:
                self._code_src = None

    def stop(self):
        self._stopped = True

    def _reset_state(self, cfg: StreamConfig):
        self._stopped = False
        now = int(time.time())
        self._run_id = f"chatcmpl_fake_{now}_{self._rng.randrange(10**6)}"
        self._msg_id = f"msg_{self._rng.randrange(10**9):09d}"
        self._chars_since_nl = 0
        self._next_nl_at = self._rng.randint(cfg.min_newline, cfg.min_newline * 4)
        self._code_pos = 0

    # ---------------------- public: async (push-friendly) ----------------------

    async def astream(self, api: APITypes = "raw", chunk: ChunkTypes = "str",
                     cps: int = 16, min_newline: int = 30,
                     include_chat_role: bool = True) -> AsyncIterator[Union[str, Dict[str, Any]]]:
        """
        Asynchronous, endless stream of chunks until stop().
        Usage:
            s = FakeOpenAIStream()
            async for payload in s.stream("chat","code",cps=20,min_newline=30):
                emit(payload)  # e.g., to a WebChannel
        """
        cfg = StreamConfig(api=api, chunk=chunk, cps=cps, min_newline=min_newline,
                           include_chat_role=include_chat_role)
        yield_from_prefix, sleep_dt = self._start_and_prefix(cfg)
        for p in yield_from_prefix:
            yield p
        try:
            while not self._stopped:
                piece = self._make_piece(cfg)
                piece = self._inject_newlines(piece, cfg.min_newline)
                yield self._wrap_payload(cfg, piece)
                await asyncio.sleep(sleep_dt)
        finally:
            # here we don't close "[DONE]" or finish_reason, because the stream is deliberately "endless"
            pass

    # ---------------------- public: sync (pull-friendly) ----------------------

    def stream(self, api: APITypes = "raw", chunk: ChunkTypes = "str",
                    cps: int = 16, min_newline: int = 30,
                    include_chat_role: bool = True) -> Iterator[Union[str, Dict[str, Any]]]:
        """
        Synchronous version (e.g., for tests in a thread without asyncio).
        """
        cfg = StreamConfig(api=api, chunk=chunk, cps=cps, min_newline=min_newline,
                           include_chat_role=include_chat_role)
        yield_from_prefix, sleep_dt = self._start_and_prefix(cfg)
        for p in yield_from_prefix:
            yield p
        try:
            while not self._stopped:
                piece = self._make_piece(cfg)
                piece = self._inject_newlines(piece, cfg.min_newline)
                yield self._wrap_payload(cfg, piece)
                time.sleep(sleep_dt)
        finally:
            pass

    # ---------------------- internals ----------------------

    def _start_and_prefix(self, cfg: StreamConfig):
        # sanity + alias
        api = (cfg.api or "raw").lower().strip()
        if api == "respones":  # alias for typo
            api = "responses"
        if api not in ("raw", "chat", "responses"):
            raise ValueError(f"Unknown api: {api}")
        chunk = (cfg.chunk or "str").lower().strip()
        if chunk not in ("str", "code"):
            raise ValueError(f"Unknown chunk: {chunk}")
        cps = max(1, min(int(cfg.cps or 16), 120))
        min_nl = max(1, int(cfg.min_newline or 30))
        self._reset_state(cfg)
        sleep_dt = 1.0 / cps

        prefix_payloads: list[Union[str, Dict[str, Any]]] = []
        # Chat: often the first delta with a role
        if api == "chat" and cfg.include_chat_role:
            prefix_payloads.append(self._wrap_chat_delta({"role": "assistant", "content": ""}))
        if chunk == "code":
            prefix_payloads.append(self._wrap_payload(cfg, ""))

        self._code_mode = (chunk == "code")
        return prefix_payloads, sleep_dt

    def _make_piece(self, cfg: StreamConfig) -> str:
        size = self._rng.randint(5, 28)
        if cfg.chunk == "code":
            if self._code_src:
                return self._code_take(size)
            return self._random_code_like(size)
        return self._random_text(size)

    def _random_text(self, n: int) -> str:
        letters = string.ascii_letters + "     "
        digits = string.digits
        punct = ".,;:!?-'" + '"'
        pool = letters + digits + punct
        return "".join(self._rng.choice(pool) for _ in range(n))

    def _random_code_like(self, n: int) -> str:
        pool = string.ascii_letters + string.digits + "_"
        ops = "()[]{}.,:;=+-*/%<>!&|^~"
        spaces_ratio = "   "
        full = pool + ops + spaces_ratio
        return "".join(self._rng.choice(full) for _ in range(n))

    def _inject_newlines(self, s: str, min_nl: int) -> str:
        if self._code_mode and self._code_src:
            return s
        out = []
        i = 0
        while i < len(s):
            remaining = self._next_nl_at - self._chars_since_nl
            if remaining <= 0:
                out.append("\n")
                self._chars_since_nl = 0
                self._next_nl_at = self._rng.randint(min_nl, min_nl * 4)
                # random indent for pseudocode
                if self._next_nl_at > min_nl and self._rng.random() < 0.25:
                    tabs = "    " * self._rng.randint(1, 3)
                    out.append(tabs)
                    self._chars_since_nl += len(tabs)
                continue
            take = min(remaining, len(s) - i)
            seg = s[i:i+take]
            out.append(seg)
            self._chars_since_nl += len(seg)
            i += take
        return "".join(out)

    def _code_take(self, n: int) -> str:
        if not self._code_src:
            return ""
        L = len(self._code_src)
        if L == 0:
            return ""
        idx = self._code_pos
        out = []
        for _ in range(n):
            out.append(self._code_src[idx])
            idx += 1
            if idx >= L:
                idx = 0
        self._code_pos = idx
        return "".join(out)

    # ---------------------- wrappers ----------------------

    def _wrap_payload(self, cfg: StreamConfig, text: str) -> Union[str, Dict[str, Any]]:
        api = "responses" if cfg.api == "respones" else cfg.api
        if api == "raw":
            return text
        elif api == "chat":
            return self._wrap_chat_delta({"content": text})
        else:
            # Responses API delta; also adding output_index/content_index/item_id
            return {
                "type": "response.output_text.delta",
                "delta": text,
                "output_index": 0,
                "content_index": 0,
                "item_id": self._msg_id
            }

    def _wrap_chat_delta(self, delta: Dict[str, Any]) -> Dict[str, Any]:
        now = int(time.time())
        return {
            "id": self._run_id,
            "object": "chat.completion.chunk",
            "created": now,
            "model": self._model,
            "choices": [
                {
                    "index": 0,
                    "delta": delta,
                    "finish_reason": None
                }
            ]
        }


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.24 23:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QVBoxLayout, QWidget

from pygpt_net.ui.widget.element.labels import TitleLabel
from pygpt_net.ui.widget.lists.mode_combo import ModeCombo
from pygpt_net.utils import trans


class Mode:
    def __init__(self, window=None):
        """
        Toolbox UI

        :param window: Window instance
        """
        self.window = window
        self.id = 'prompt.mode'
        self._label_key = f'{self.id}.label'
        self._label_trans_key = 'toolbox.mode.label'

    def setup(self) -> QWidget:
        """
        Setup mode

        :return: QWidget
        """
        widget = QWidget()
        widget.setLayout(self.setup_list())
        return widget

    def setup_list(self) -> QVBoxLayout:
        """
        Setup list

        :return: QVBoxLayout
        """
        ui_nodes = self.window.ui.nodes

        label = ui_nodes.get(self._label_key)
        label_text = trans(self._label_trans_key)
        if label is None:
            label = TitleLabel(label_text)
            ui_nodes[self._label_key] = label
        else:
            label.setText(label_text)

        combo = ui_nodes.get(self.id)
        if combo is None:
            combo = ModeCombo(self.window, self.id)
            ui_nodes[self.id] = combo

        header_layout = QVBoxLayout()
        header_layout.addWidget(label)
        header_layout.setContentsMargins(5, 5, 0, 0)

        layout = QVBoxLayout()
        layout.addLayout(header_layout)
        layout.addWidget(combo)
        layout.addStretch()
        layout.setContentsMargins(2, 5, 5, 5)

        return layout


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.15 22:00:00                  #
# ================================================== #

import copy
import datetime
import uuid
from typing import Optional, Tuple, List, Dict

from packaging.version import Version

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_ASSISTANT,
    MODE_AUDIO,
    MODE_CHAT,
    MODE_COMPLETION,
    MODE_EXPERT,
    MODE_IMAGE,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
    MODE_VISION,
    MODE_RESEARCH,
    MODE_COMPUTER,
    MODE_AGENT_OPENAI,
)
from pygpt_net.item.ctx import CtxItem, CtxMeta, CtxGroup
from pygpt_net.provider.core.ctx.base import BaseProvider
from pygpt_net.provider.core.ctx.db_sqlite import DbSqliteProvider
from pygpt_net.utils import trans

from .idx import Idx
from .container import Container
from .output import Output


class Ctx:
    def __init__(self, window=None):
        """
        Context core

        :param window: Window instance
        """
        self.window = window
        self.provider = DbSqliteProvider(window)
        self.container = Container(window)  # context container
        self.output = Output(window)  # context render output
        self.idx = Idx(window)  # context indexing core
        self.meta = {}
        self.current = None
        self.last_item = None
        self.assistant = None
        self.mode = None
        self.model = None
        self.preset = None
        self.run = None
        self.status = None
        self.thread = None
        self.last_mode = None
        self.last_model = None
        self.tmp_meta = None
        self.search_string = None  # search string
        self.groups = {}  # groups
        self.filters = {}  # search filters
        self.filters_labels = []  # search labels
        self.current_cmd = []  # current commands
        self.current_cmd_schema = "" # current commands schema
        self.all_modes = [
            MODE_CHAT,
            MODE_COMPLETION,
            MODE_IMAGE,
            MODE_LANGCHAIN,
            MODE_VISION,
            MODE_ASSISTANT,
            MODE_LLAMA_INDEX,
            MODE_AGENT,
            MODE_AGENT_LLAMA,
            MODE_AGENT_OPENAI,
            MODE_EXPERT,
            MODE_AUDIO,
            MODE_RESEARCH,
            MODE_COMPUTER,
        ]
        self.allowed_modes = {
            MODE_CHAT: self.all_modes,
            MODE_COMPLETION: self.all_modes,
            MODE_IMAGE: [MODE_IMAGE],
            MODE_LANGCHAIN: self.all_modes,
            MODE_VISION: self.all_modes,
            MODE_ASSISTANT: [MODE_ASSISTANT],
            MODE_LLAMA_INDEX: self.all_modes,
            MODE_EXPERT: self.all_modes,
            MODE_AUDIO: self.all_modes,
            MODE_AGENT: self.all_modes,
            MODE_AGENT_LLAMA: self.all_modes,
            MODE_AGENT_OPENAI: self.all_modes,
            MODE_RESEARCH: self.all_modes,
            MODE_COMPUTER: self.all_modes,
        }
        self.current_sys_prompt = ""
        self.groups_loaded = False

    def get_items(self) -> List[CtxItem]:
        """
        Get context items

        :return: context items
        """
        return self.container.get_items()

    def set_items(self, items: List[CtxItem]):
        """
        Set context items

        :param items: context items
        """
        self.container.set_items(items)

    def clear_items(self):
        """Clear context items"""
        self.container.clear_items()

    def count_items(self) -> int:
        """
        Count context items
        :return: context items count
        """
        return self.container.count_items()

    def get_current(self) -> int:
        """
        Get current context ID

        :return: current context ID
        """
        return self.current

    def set_current(self, current: int):
        """
        Set current context ID

        :param current: current context ID
        """
        self.current = current

    def clear_current(self):
        """Clear current context ID"""
        self.current = None

    def get_last_item(self) -> Optional[CtxItem]:
        """
        Get last item

        :return: last item
        """
        return self.last_item

    def set_last_item(self, last_item: Optional[CtxItem]):
        """
        Set last item

        :param last_item: last item
        """
        self.last_item = last_item

    def get_assistant(self) -> str:
        """
        Get assistant name

        :return: assistant name
        """
        return self.assistant

    def set_assistant(self, assistant: str):
        """
        Set assistant name

        :param assistant: assistant name
        :return: assistant name
        """
        self.assistant = assistant

    def get_mode(self) -> str:
        """
        Get mode

        :return: mode
        """
        return self.mode

    def set_mode(self, mode: str):
        """
        Set mode

        :param mode: mode
        """
        self.mode = mode

    def get_model(self) -> str:
        """
        Get model name
        :return: model name
        """
        return self.model

    def set_model(self, model: str):
        """
        Set model name

        :param model: model name
        """
        self.model = model

    def get_preset(self) -> str:
        """
        Get preset name
        :return: preset
        """
        return self.preset

    def set_preset(self, preset: str):
        """
        Set preset name

        :param preset: preset name
        """
        self.preset = preset

    def get_run(self) -> str:
        """
        Get run ID

        :return: run ID
        """
        return self.run

    def set_status(self, status: int):
        """
        Set status (label color)

        :param status: status
        """
        self.status = status

    def get_status(self) -> int:
        """
        Get status (label color)

        :return: status
        """
        return self.status

    def set_run(self, run: str):
        """
        Set run ID

        :param run: run ID
        """
        self.run = run

    def get_thread(self) -> str:
        """
        Get thread ID

        :return: thread ID
        """
        return self.thread

    def set_thread(self, thread: str):
        """
        Set thread ID

        :param thread: thread ID
        """
        self.thread = thread

    def clear_thread(self):
        """Clear thread ID"""
        self.thread = None

    def get_last_mode(self) -> str:
        """
        Get last mode

        :return: last mode
        """
        return self.last_mode

    def set_last_mode(self, last_mode: str):
        """
        Set last mode

        :param last_mode: last mode
        """
        self.last_mode = last_mode

    def get_last_model(self) -> str:
        """
        Get last model name

        :return: last model
        """
        return self.last_model

    def set_last_model(self, last_model: str):
        """
        Set last model name

        :param last_model: last model name
        """
        self.last_model = last_model

    def get_tmp_meta(self) -> Optional[CtxMeta]:
        """
        Get temporary meta

        :return: temporary meta
        """
        return self.tmp_meta

    def set_tmp_meta(self, tmp_meta: Optional[CtxMeta]):
        """
        Set temporary meta

        :param tmp_meta: temporary meta
        """
        self.tmp_meta = tmp_meta

    def get_search_string(self) -> Optional[str]:
        """
        Get search string

        :return: search string
        """
        return self.search_string

    def set_search_string(self, search_string: Optional[str]):
        """
        Set search string

        :param search_string: search string
        """
        self.search_string = search_string

    def clear_search_string(self):
        """Clear search string"""
        self.search_string = None

    def install(self):
        """Install provider data"""
        self.provider.install()

    def patch(self, app_version: Version) -> bool:
        """
        Patch provider data

        :param app_version: app version
        :return: True if data was patched
        """
        return self.provider.patch(app_version)

    def get_provider(self) -> BaseProvider:
        """
        Get provider instance

        :return: provider instance
        """
        return self.provider

    def select(
            self,
            id: int,
            restore_model: bool = True
    ):
        """
        Select ctx meta by ID and load ctx items

        :param id: context meta id
        :param restore_model: restore model
        """
        if id not in self.meta:
            self.load_tmp_meta(id)

        if id in self.meta:
            ctx = self.meta[id]
            self.current = id

            self.thread = None
            self.mode = None
            self.assistant = None

            self.thread = ctx.thread
            self.mode = ctx.mode
            self.assistant = ctx.assistant
            self.preset = ctx.preset

            if restore_model:
                if ctx.last_model is not None \
                        and self.window.core.models.has_model(self.mode, ctx.last_model):
                    self.model = ctx.last_model
                elif ctx.model is not None \
                        and self.window.core.models.has_model(self.mode, ctx.model):
                    self.model = ctx.model

            self.set_items(self.load(id))

    def new(
            self,
            group_id: Optional[int] = None
    ) -> Optional[CtxMeta]:
        """
        Create new ctx and set as current

        :param group_id: group id
        :return: CtxMeta instance (new ctx meta)
        """
        meta = self.create(group_id)
        if meta is None:
            self.window.core.debug.log("Error creating new ctx")
            return
        preset = self.window.core.config.get('preset')
        meta.preset = preset
        self.meta[meta.id] = meta
        self.tmp_meta = meta
        self.current = meta.id
        self.last_item = None
        self.thread = None
        self.assistant = None
        self.mode = self.window.core.config.get('mode')
        self.model = self.window.core.config.get('model')
        self.preset = preset
        self.clear_items()
        self.save(meta.id)

        return meta

    def build(self) -> CtxMeta:
        """
        Build new ctx

        :return: created CtxMeta instance
        """
        meta = CtxMeta()
        meta.name = "{}".format(trans('ctx.new.prefix'))
        meta.date = datetime.datetime.now().strftime("%Y-%m-%d")
        meta.mode = self.window.core.config.get('mode')
        meta.model = self.window.core.config.get('model')
        meta.last_mode = self.window.core.config.get('mode')
        meta.last_model = self.window.core.config.get('model')
        meta.initialized = False
        return meta

    def create(
            self,
            group_id: Optional[int] = None
    ) -> CtxMeta:
        """
        Send created meta to provider and return new ID

        :param group_id: group id
        :return: CtxMeta instance
        """
        meta = self.build()
        if group_id is not None:
            meta.group_id = group_id
        id = self.provider.create(meta)
        meta.id = id
        return meta

    def add(
            self,
            item: CtxItem,
            parent_id: Optional[int] = None
    ):
        """
        Add CtxItem to contexts and saves context

        :param item: CtxItem to append
        :param parent_id: parent id
        """
        if parent_id is not None:
            self.add_to_meta(item, parent_id)
            return

        items = self.get_items()
        items.append(item)

        if self.current is not None:
            if self.current not in self.meta:
                self.load_tmp_meta(self.current)
            if self.current in self.meta:
                meta = self.meta[self.current]
                result = self.provider.append_item(meta, item)
                if not result:
                    self.store()

    def add_to_meta(
            self,
            item: CtxItem,
            meta_id: Optional[int] = None
    ):
        """
        Add CtxItem to custom meta

        :param item: CtxItem to append
        :param meta_id: meta id
        """
        if meta_id in self.meta:
            meta = self.meta[meta_id]
        else:
            meta = self.provider.get_meta_by_id(meta_id)
        if meta is not None:
            self.provider.append_item(meta, item)

    def update_item(self, item: CtxItem):
        """
        Update CtxItem in context

        :param item: CtxItem to update
        """
        self.provider.update_item(item)

    def update_indexed_ts_by_id(self, id: int, ts: int):
        """
        Update indexed timestamp by ID

        :param id: context id
        :param ts: timestamp
        """
        if id in self.meta:
            self.meta[id].indexed = ts

    def is_empty(self) -> bool:
        """
        Check if ctx is empty

        :return: True if empty, false otherwise
        """
        if self.current is None:
            return True
        return self.count_items() == 0

    def update(self):
        """
        Update current parent (when ctx item load from the list or setting mode)
        """
        self.mode = self.window.core.config.get('mode')

        if self.current is None:
            return

        if self.current not in self.meta:
            self.load_tmp_meta(self.current)

        if self.current not in self.meta:
            return

        self.meta[self.current].mode = self.mode
        self.save(self.current)

    def replace(self, meta: CtxMeta):
        """
        Replace meta

        :param meta: CtxMeta
        """
        self.meta[meta.id] = meta

    def post_update(self, mode: str):
        """
        Update current (last) ctx data

        :param mode: mode name
        """
        if self.current is None:
            return

        if self.current not in self.meta:
            self.load_tmp_meta(self.current)

        if self.current not in self.meta:
            return

        self.assistant = self.window.core.config.get('assistant')
        self.preset = self.window.core.config.get('preset')
        self.model = self.window.core.config.get('model')

        self.meta[self.current].last_mode = mode
        self.meta[self.current].last_model = self.model
        self.meta[self.current].preset = self.preset

        if mode == MODE_ASSISTANT:
            self.meta[self.current].assistant = self.assistant

        self.save(self.current)

    def is_initialized(self) -> bool:
        """
        Check if ctx is initialized (name assigned)

        :return: True if initialized, false otherwise
        """
        if self.current is None:
            return False
        if self.current not in self.meta:
            self.load_tmp_meta(self.current)
        if self.current in self.meta:
            return self.meta[self.current].initialized
        return False

    def set_initialized(self):
        """Set ctx as initialized (name assigned)"""
        if self.current is None:
            return
        if self.current not in self.meta:
            self.load_tmp_meta(self.current)
        if self.current in self.meta:
            self.meta[self.current].initialized = True
            self.save(self.current)

    def has(self, id: int) -> bool:
        """
        Check if ctx meta exists

        :param id: ctx ID
        :return: True if exists, false otherwise
        """
        return id in self.meta

    def get(self, idx: int) -> CtxItem:
        """
        Return ctx item by index

        :param idx: item index
        :return: context item
        """
        if idx < self.count_items():
            return self.get_items()[idx]

    def get_item_by_id(self, id: int) -> CtxItem:
        """
        Return ctx item by id

        :param id: item id
        :return: context item
        """
        items = self.get_items()
        for item in items:
            if item.id == id:
                return item
        return self.fetch_item_by_id(id)

    def fetch_item_by_id(self, id: int) -> CtxItem:
        """
        Fetch ctx item by id

        :param id: item id
        :return: context item
        """
        return self.provider.get_item_by_id(id)

    def get_meta(self, reload: bool = False) -> dict:
        """
        Get ctx items sorted descending by date

        :param reload: True if reload from provider
        :return: ctx metas dict
        """
        if reload:
            self.load_meta()
        return self.meta

    def get_current_meta(self) -> Optional[CtxMeta]:
        """
        Get current meta

        :return: current meta
        """
        if self.current is not None:
            return self.get_meta_by_id(self.current)

    def get_id_by_idx(self, idx: int) -> int:
        """
        Get ctx id (id) by index

        :param idx: index
        :return: ctx id
        """
        i = 0
        for id in self.get_meta():
            if i == idx:
                return id
            i += 1

    def get_idx_by_id(self, id: int) -> int:
        """
        Get ctx index by id

        :param id: id
        :return: idx
        """
        for i, key in enumerate(self.get_meta()):
            if key == id:
                return i

    def get_first(self) -> str:
        """
        Return first ctx ID from list

        :return: ctx id
        """
        for id in self.get_meta():
            return id

    def get_meta_by_id(self, id: int) -> CtxMeta:
        """
        Return ctx meta by id

        :param id: ctx id
        :return: ctx meta object
        """
        if id in self.meta:
            return self.meta[id]
        else:
            self.load_tmp_meta(id)

    def get_last(self) -> Optional[CtxItem]:
        """
        Return last item from ctx

        :return: last ctx item
        """
        items = self.get_items()
        if items:
            return items[-1]
        return None

    def is_first_item(self, item_id: int) -> bool:
        """
        Check if item is first in ctx

        :param item_id: item id
        :return: True if first
        """
        items = self.get_items()
        if items:
            return items[0].id == item_id
        return False

    def is_last_item(self, item_id: int) -> bool:
        """
        Check if item is last in ctx

        :param item_id: item id
        :return: True if last
        """
        items = self.get_items()
        if items:
            return items[-1].id == item_id
        return False

    def get_previous_item(self, item_id: int) -> Optional[CtxItem]:
        """
        Get previous item from ctx

        :param item_id: item id
        :return: ctx item
        """
        items = self.get_items()
        prev = None
        for it in items:
            if it.id == item_id:
                return prev
            prev = it
        return None

    def prepare(self):
        """Prepare context for prompt"""
        if self.count_meta() == 0:
            self.new()

    def count(self) -> int:
        """
        Count ctx items

        :return: ctx items count
        """
        return self.count_items()

    def count_meta(self) -> int:
        """
        Count ctx meta items

        :return: ctx meta count
        """
        return len(self.meta) + (1 if self.tmp_meta is not None else 0)

    def count_found_meta(self) -> int:
        """
        Count ctx meta items

        :return: ctx meta count
        """
        return len(self.meta)

    def all(
            self,
            meta_id: Optional[int] = None
    ) -> List[CtxItem]:
        """
        Return ctx items (current or by meta_id if provided)

        :param meta_id: meta id
        :return: ctx items
        """
        if meta_id is None:
            return self.get_items()
        else:
            return self.load(meta_id)

    def remove(self, id: int):
        """
        Delete ctx by id

        :param id: ctx id
        """
        if id in self.meta:
            del self.meta[id]
            self.provider.remove(id)

    def remove_item(self, id: int):
        """
        Remove ctx item by id

        :param id: ctx id
        """
        items = self.get_items()
        for i, item in enumerate(items):
            if item.id == id:
                items.pop(i)
                self.provider.remove_item(id)
                break

    def remove_items_from(
            self,
            meta_id: int,
            item_id: int
    ):
        """
        Remove ctx items from meta_id

        :param meta_id: meta_id
        :param item_id: item_id
        """
        items = [item for item in self.get_items() if item.id < item_id]
        self.set_items(items)
        return self.provider.remove_items_from(meta_id, item_id)

    def truncate(self):
        """Delete all ctx"""
        self.meta = {}
        self.provider.truncate()

    def clear(self):
        """Clear ctx items"""
        self.clear_items()

    def append_thread(self, thread: str):
        """
        Append thread ID to ctx

        :param thread: thread ID
        """
        self.thread = thread
        if self.current is None:
            return
        if self.current not in self.meta:
            self.load_tmp_meta(self.current)
        if self.current in self.meta:
            self.meta[self.current].thread = self.thread
            self.save(self.current)

    def append_run(self, run):
        """
        Append run ID to ctx

        :param run: run ID
        """
        self.run = run
        if self.current is None:
            return
        if self.current not in self.meta:
            self.load_tmp_meta(self.current)
        if self.current in self.meta:
            self.meta[self.current].run = self.run
            self.save(self.current)

    def append_status(self, status: int):
        """
        Append status (label color) to ctx

        :param status: status
        """
        self.status = status
        if self.current is None:
            return
        if self.current in self.meta:
            self.meta[self.current].status = self.status
            self.save(self.current)

    def get_or_create_slave_meta(
            self,
            master_ctx: CtxItem,
            preset_id: str
    ) -> CtxMeta:
        """
        Get or create slave meta

        :param master_ctx: master context
        :param preset_id: preset ID
        :return: slave meta
        """
        slaves = []
        if master_ctx.meta is not None:
            slaves = self.provider.get_meta_by_root_id_and_preset_id(
                master_ctx.meta.id,
                preset_id,
            )
        if len(slaves) > 0:
            return next(iter(slaves.values()))
        slave = self.build()

        if master_ctx.meta is not None:
            slave.root_id = master_ctx.meta.id
            slave.parent_id = master_ctx.meta.id

        slave.preset = preset_id
        id = self.provider.create(slave)
        slave.id = id
        return slave

    def get_prev(self) -> Optional[int]:
        """
        Get previous context

        :return: previous context
        """
        if self.current is None:
            return None
        idx = self.get_idx_by_id(self.current)
        if idx is not None and idx > 0:
            return self.get_id_by_idx(idx - 1)

    def get_next(self) -> Optional[int]:
        """
        Get next context

        :return: next context
        """
        if self.current is None:
            return None
        idx = self.get_idx_by_id(self.current)
        if idx is not None and idx < self.count_meta() - 1:
            return self.get_id_by_idx(idx + 1)

    def get_last_meta(self) -> Optional[int]:
        """
        Get last context

        :return: last meta
        """
        return self.provider.get_last_meta_id()

    def count_history(
            self,
            history_items: List[CtxItem],
            model: str,
            mode: str,
            used_tokens: int = 100,
            max_tokens: int = 1000
    ) -> Tuple[int, int]:
        """
        Count ctx items to add to prompt

        :param history_items: history items list
        :param model: model
        :param mode: mode
        :param used_tokens: used tokens
        :param max_tokens: max tokens
        :return: ctx items count, ctx tokens count
        """
        i = 0
        tokens = used_tokens
        context_tokens = 0
        from_ctx = self.window.core.tokens.from_ctx
        for item in reversed(history_items):
            num = from_ctx(item, mode, model)
            new_total = tokens + num
            if 0 < max_tokens < new_total:
                break
            tokens = new_total
            context_tokens += num
            i += 1

        return i, context_tokens

    def get_history(
            self,
            history_items: List[CtxItem],
            model: str,
            mode: str = MODE_CHAT,
            used_tokens: int = 100,
            max_tokens: int = 1000,
            ignore_first: bool = True
    ) -> list:
        """
        Return ctx items to add to prompt

        :param history_items: history items list
        :param model: model
        :param mode: mode
        :param used_tokens: used tokens
        :param max_tokens: max tokens
        :param ignore_first: ignore current item (provided by user)
        :return: ctx items list
        """
        items = []
        tokens = used_tokens
        is_first = True
        from_ctx = self.window.core.tokens.from_ctx
        for item in reversed(history_items):
            if is_first and ignore_first:
                is_first = False
                continue
            cost = from_ctx(item, mode, model)
            new_total = tokens + cost
            if 0 < max_tokens < new_total:
                break
            tokens = new_total
            items.append(item)

        items.reverse()
        return items

    def count_prompt_items(
            self,
            model: str,
            mode: str,
            used_tokens: int = 100,
            max_tokens: int = 1000
    ) -> Tuple[int, int]:
        """
        Count ctx items to add to prompt

        :param model: model
        :param mode: mode
        :param used_tokens: used tokens
        :param max_tokens: max tokens
        :return: ctx items count, ctx tokens count
        """
        return self.count_history(
            self.get_items(),
            model,
            mode,
            used_tokens,
            max_tokens,
        )

    def get_prompt_items(
            self,
            model: str,
            mode: str = MODE_CHAT,
            used_tokens: int = 100,
            max_tokens: int = 1000,
            ignore_first: bool = True
    ) -> list:
        """
        Return ctx items to add to prompt

        :param model: model
        :param mode: mode
        :param used_tokens: used tokens
        :param max_tokens: max tokens
        :param ignore_first: ignore current item (provided by user)
        :return: ctx items list
        """
        return self.get_history(
            self.get_items(),
            model,
            mode,
            used_tokens,
            max_tokens,
            ignore_first,
        )

    def get_all_items(self, ignore_first: bool = True) -> list:
        """
        Return all ctx items

        :param ignore_first: ignore current item (provided by user)
        :return: ctx items list
        """
        items = self.get_items()
        if ignore_first:
            return items[:-1]
        return items[:]

    def check(self, threshold: int, max_total: int):
        """
        Check context and clear if limit exceeded

        :param threshold: threshold
        :param max_total: max total tokens
        """
        if self.get_tokens_left(max_total) <= threshold:
            self.remove_first()

    def get_tokens_left(self, max: int) -> int:
        """
        Return remaining tokens in context

        :param max: max tokens
        :return: remaining tokens in context
        """
        return max - self.get_total_tokens()

    def get_total_tokens(self) -> int:
        """
        Return current prompt tokens

        :return: total tokens
        """
        last = self.get_last()
        if last is not None:
            return last.total_tokens
        return 0

    def get_last_tokens(self) -> int:
        """
        Return last tokens count

        :return: last tokens
        """
        return self.get_total_tokens()

    def remove_last(self):
        """Remove last item"""
        items = self.get_items()
        if items:
            items.pop()

    def duplicate(self, id: int) -> int:
        """
        Duplicate ctx and return new ctx id

        :param id: ctx id
        :return: new ctx id
        """
        if id in self.meta:
            meta = self.create(self.meta[id].group_id)
            new_id = meta.id
            old_data = self.meta[id].to_dict()
            meta.from_dict(old_data)
            meta.id = new_id
            items = self.load(id)
            self.provider.save_all(meta.id, meta, items)
            return meta.id

    def remove_first(self):
        """Remove first item"""
        items = self.get_items()
        if items:
            items.pop(0)

    def set_display_filters(self, filters: dict):
        """
        Update current display filters

        :param filters: filters dict
        """
        self.filters = filters

    def is_allowed_for_mode(
            self,
            mode: str,
            check_assistant: bool = True
    ) -> bool:
        """
        Check if ctx is allowed for this mode

        :param mode: mode name
        :param check_assistant: True if check also current assistant
        :return: True if allowed for mode
        """
        if not self.window.core.config.get('lock_modes'):
            return True

        if self.is_empty():
            return True

        if self.current is None or self.current == '' or not self.has(self.current):
            return True

        meta = self.get_meta_by_id(self.current)

        if meta.last_mode is None:
            return True

        prev_mode = meta.last_mode
        if prev_mode not in self.allowed_modes[mode]:
            if mode == MODE_ASSISTANT:
                if meta.assistant is not None:
                    if meta.assistant == self.window.core.config.get('assistant'):
                        return True
                else:
                    return True
            return False

        if mode == MODE_ASSISTANT and check_assistant:
            if meta.assistant is None:
                return True
            if meta.assistant != self.window.core.config.get('assistant'):
                return False
        return True

    def has_labels(self) -> bool:
        """
        Check if label query is needed

        :return: True if labels not default
        """
        num_all = len(self.window.controller.ui.get_colors())
        return len(self.filters_labels) < num_all

    def load_meta(self):
        """Load ctx list: pinned and grouped unlimited; ungrouped not pinned paginated directly in SQL."""
        # base package size (per page)
        base_limit = 0
        if self.window.core.config.has('ctx.records.limit'):
            try:
                base_limit = int(self.window.core.config.get('ctx.records.limit') or 0)
            except Exception:
                base_limit = 0

        # total loaded (persisted); fallback to base when missing
        try:
            loaded_total = int(self.window.core.config.get('ctx.records.limit.total') or 0)
        except Exception:
            loaded_total = 0
        if loaded_total <= 0:
            loaded_total = base_limit

        # Common filters (labels etc.)
        common_filters = self.get_parsed_filters()

        # If explicit filters target a narrow subset (legacy path), keep old behavior
        if "is_important" in self.filters or "indexed_ts" in self.filters:
            limit = 0 if base_limit == 0 else loaded_total
            self.meta = self.provider.get_meta(
                search_string=self.search_string,
                order_by='updated_ts',
                order_direction='DESC',
                limit=limit,
                offset=0,
                filters=common_filters,
                search_content=self.is_search_content(),
            )
            return

        # 1) Pinned (important) – unlimited
        filters_pinned = self.get_parsed_filters()
        filters_pinned['is_important'] = {"mode": "=", "value": 1}
        meta_pinned = self.provider.get_meta(
            search_string=self.search_string,
            order_by='updated_ts',
            order_direction='DESC',
            limit=0,
            offset=0,
            filters=filters_pinned,
            search_content=self.is_search_content(),
        )

        # 2) Grouped – unlimited
        filters_grouped = self.get_parsed_filters()
        filters_grouped['group_id'] = {"mode": ">", "value": 0}
        meta_grouped = self.provider.get_meta(
            search_string=self.search_string,
            order_by='updated_ts',
            order_direction='DESC',
            limit=0,
            offset=0,
            filters=filters_grouped,
            search_content=self.is_search_content(),
        )

        # 3) Ungrouped & not pinned – paginate directly in SQL
        #    If base_limit == 0 -> unlimited (no paging)
        filters_ungrp = self.get_parsed_filters()
        filters_ungrp['is_important'] = {"mode": "=", "value": 0}
        filters_ungrp['group_id'] = {"mode": "NULL_OR_ZERO", "value": 0}  # special mode handled in Storage

        if base_limit <= 0:
            meta_ungrouped = self.provider.get_meta(
                search_string=self.search_string,
                order_by='updated_ts',
                order_direction='DESC',
                limit=0,         # unlimited
                offset=0,
                filters=filters_ungrp,
                search_content=self.is_search_content(),
            )
        else:
            # Always take the top-N ungrouped newest items directly from DB
            take = max(0, int(loaded_total or 0))
            meta_ungrouped = self.provider.get_meta(
                search_string=self.search_string,
                order_by='updated_ts',
                order_direction='DESC',
                limit=take,
                offset=0,
                filters=filters_ungrp,
                search_content=self.is_search_content(),
            )

        # Compose final dict with deterministic order: pinned -> grouped -> ungrouped
        combined = {}
        combined.update(meta_pinned)
        combined.update(meta_grouped)
        combined.update(meta_ungrouped)
        self.meta = combined

    def load_tmp_meta(self, meta_id: int):
        """
        Load tmp meta

        :param meta_id: meta id
        """
        meta = self.provider.get_meta(
            order_by='updated_ts',
            order_direction='DESC',
            limit=1,
            filters={
                "id": {
                    "mode": "=",
                    "value": meta_id,
                },
            }
        )
        if len(meta) > 0:
            self.tmp_meta = next(iter(meta.values()))
            if self.tmp_meta.id not in self.meta:
                self.meta = {self.tmp_meta.id: self.tmp_meta, **self.meta}

    def clear_tmp_meta(self):
        """Clear tmp meta"""
        if self.tmp_meta is not None and self.current != self.tmp_meta.id:
            self.tmp_meta = None

    def get_parsed_filters(self) -> dict:
        """
        Get parsed filters

        :return: parsed filters
        """
        filters = copy.deepcopy(self.filters)
        if self.has_labels():
            filters['label'] = {
                "mode": "IN",
                "value": self.filters_labels,
            }
        return filters

    def load(self, id: int) -> List[CtxItem]:
        """
        Load ctx items from provider and append meta to each item

        :param id: ctx id
        :return: ctx items list
        """
        items = self.provider.load(id)
        meta = self.get_meta_by_id(id)
        for item in items:
            item.meta = meta
        return items

    def load_groups(self):
        """Load groups"""
        self.groups = self.provider.get_groups()

    def has_group(self, id: int) -> bool:
        """
        Check if group exists

        :param id: group id
        :return: True if exists
        """
        return id in self.groups

    def get_groups(self) -> Dict[int, CtxGroup]:
        """
        Get groups

        :return: groups
        """
        if not self.groups_loaded:
            self.load_groups()
            self.groups_loaded = True
        return self.groups

    def get_group_by_id(self, id: int) -> Optional[CtxGroup]:
        """
        Get group by ID

        :param id: group id
        :return: group instance
        """
        if not self.groups_loaded:
            self.load_groups()
            self.groups_loaded = True
        if id not in self.groups:
            return None
        return self.groups[id]

    def update_model_in_current(self, model):
        """
        Update model in current context

        :param model: model name
        """
        if self.current is None:
            return
        meta = self.meta.get(self.current)
        if meta is not None:
            meta.last_model = model
            self.save(self.current)

    def remove_group(
            self,
            group: CtxGroup,
            all: bool = False
    ):
        """
        Remove group

        :param group: group instance
        :param all: remove all items
        """
        self.provider.remove_group(group.id, all=all)
        self.load_groups()

    def truncate_groups(self):
        """Remove all groups"""
        self.provider.truncate_groups()
        self.load_groups()

    def make_group(self, name: str) -> CtxGroup:
        """
        Make group

        :param name: group name
        :return: group instance
        """
        group = CtxGroup()
        group.uuid = str(uuid.uuid4())
        group.name = name
        return group

    def insert_group(self, group: CtxGroup) -> int:
        """
        Insert group

        :param group: group instance
        :return: group id
        """
        id = self.provider.insert_group(group)
        self.load_groups()
        return id

    def update_group(self, group: CtxGroup):
        """
        Update group

        :param group: group instance
        """
        self.provider.update_group(group)
        self.load_groups()

    def update_meta_group_id(self, id: int, group_id: int):
        """
        Update meta group ID

        :param id: meta id
        :param group_id: group id
        """
        meta = self.get_meta_by_id(id)
        if meta is None:
            return
        meta.group_id = group_id
        self.provider.update_meta_group_id(id, group_id)
        self.load_groups()

    def get_items_by_id(self, id: int) -> List[str]:
        """
        Get ctx items by id as string list

        :param id: ctx id
        :return: ctx items list
        """
        items = self.provider.load(id)
        data = []
        for item in items:
            data.append("Human: " + str(item.input) + "\n" + "Assistant: " + str(item.output) + "\n")
        return data

    def get_list_in_date_range(
            self,
            search_string: Optional[str] = None,
            limit: int = 0
    ) -> List[Dict[str, Dict]]:
        """
        Get ctx list in date range

        :param search_string: search string
        :param limit: limit
        :return: ctx list
        """
        meta = self.provider.get_meta(
            search_string=search_string,
            order_by='updated_ts',
            order_direction='DESC',
            limit=limit,
            filters={},
            search_content=self.is_search_content(),
        )
        data = []
        for key in meta:
            item = meta[key]
            data.append({key: {
                "subject": item.name,
                "last_updated": datetime.datetime.fromtimestamp(item.updated).strftime('%Y-%m-%d %H:%M:%S'),
            }})
        return data

    def is_search_content(self) -> bool:
        """
        Check if search in content is enabled

        :return: True if enabled
        """
        return bool(self.window.core.config.get('ctx.search_content'))

    def save(self, id: int):
        """
        Save ctx data

        :param id: ctx id
        """
        self.provider.save(id, self.meta[id], self.get_items())

    def store(self):
        """Store current ctx"""
        cur = self.current
        if cur is not None:
            if cur not in self.meta:
                self.load_tmp_meta(cur)
            if cur in self.meta:
                self.save(cur)

    def reset_meta(self, id: int):
        """
        Reset meta

        :param id: meta id
        """
        if id in self.meta:
            self.provider.clear_meta(id)
            self.meta[id].initialized = False
        self.load_meta()

    def as_previous(self, ctx: CtxItem) -> CtxItem:
        """
        Prepare previous context item and clear current reply results

        :param ctx: CtxItem instance (current)
        :return: CtxItem instance (previous)
        """
        prev_ctx = CtxItem()
        for name in (
            "urls", "urls_before", "images", "images_before", "files", "files_before",
            "attachments", "attachments_before", "results", "index_meta", "doc_ids",
            "input_name", "output_name"
        ):
            setattr(prev_ctx, name, copy.deepcopy(getattr(ctx, name)))
        ctx.clear_reply()
        if len(ctx.cmds) == 0:
            ctx.from_previous()
        return prev_ctx

    def dump(self, ctx: CtxItem) -> str:
        """
        Dump context item

        :param ctx: CtxItem instance
        """
        return self.provider.dump(ctx)

    def reset(self):
        """Reset all data"""
        self.meta = {}
        self.clear_items()
        self.current = None
        self.last_item = None
        self.assistant = None
        self.mode = None
        self.model = None
        self.preset = None
        self.run = None
        self.status = None
        self.thread = None
        self.last_mode = None
        self.last_model = None
        self.tmp_meta = None
        self.search_string = None  # search string
        self.groups = {}  # groups
        self.filters = {}  # search filters
        self.filters_labels = []  # search labels
        self.current_cmd = []  # current commands
        self.current_cmd_schema = ""  # current commands schema
        self.current_sys_prompt = ""
        self.groups_loaded = False


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.05 17:00:00                  #
# ================================================== #

from PySide6.QtGui import QAction, QIcon

from pygpt_net.utils import trans

class Tools:
    def __init__(self, window=None):
        """
        Tools menu setup

        :param window: Window instance
        """
        self.window = window

    def _open_tab_action(self, checked=False):
        action = self.window.sender()
        self.window.controller.tools.open_tab(action.data())

    def _toggle_remote_store(self, checked=False):
        self.window.controller.remote_store.toggle_editor()

    def _rebuild_ipython(self, checked=False):
        self.window.core.plugins.get("cmd_code_interpreter").builder.build_and_restart()

    def _rebuild_python_legacy(self, checked=False):
        self.window.core.plugins.get("cmd_code_interpreter").docker.build_and_restart()

    def _rebuild_system(self, checked=False):
        self.window.core.plugins.get("cmd_system").docker.build_and_restart()

    def setup(self):
        """Setup tools menu"""
        window = self.window
        ui_menu = window.ui.menu

        tab_tools = window.controller.tools.get_tab_tools()
        ui_menu['menu.tools'] = window.menuBar().addMenu(trans("menu.tools"))
        menu_tools = ui_menu['menu.tools']

        for key, val in tab_tools.items():
            label_key, icon_name, type_ = val[0], val[1], val[2]
            action = QAction(QIcon(f":/icons/{icon_name}.svg"), trans(f"output.tab.{label_key}"), window)
            action.setCheckable(False)
            action.triggered.connect(lambda checked=False, t=type_: window.controller.tools.open_tab(t))
            ui_menu[key] = action
            menu_tools.addAction(action)

        actions = window.tools.setup_menu_actions()
        if len(actions) == 0:
            return

        menu_tools.addSeparator()

        for key, action in actions.items():
            ui_menu[key] = action
            menu_tools.addAction(action)

        # ------------------------------------------------- #

        menu_tools.addSeparator()
        db_icon = QIcon(":/icons/db.svg")
        ui_menu['menu.tools.remote_store'] = QAction(db_icon, trans("dialog.remote_store"), window)
        menu_tools.addAction(ui_menu['menu.tools.remote_store'])
        ui_menu['menu.tools.remote_store'].triggered.connect(self._toggle_remote_store)
        # ------------------------------------------------- #

        menu_tools.addSeparator()
        reload_icon = QIcon(":/icons/reload.svg")
        ui_menu['menu.tools.ipython.rebuild'] = QAction(reload_icon, "Rebuild IPython Docker Image", window)
        menu_tools.addAction(ui_menu['menu.tools.ipython.rebuild'])
        ui_menu['menu.tools.ipython.rebuild'].triggered.connect(self._rebuild_ipython)

        ui_menu['menu.tools.python_legacy.rebuild'] = QAction(reload_icon, "Rebuild Python (Legacy) Docker Image", window)
        menu_tools.addAction(ui_menu['menu.tools.python_legacy.rebuild'])
        ui_menu['menu.tools.python_legacy.rebuild'].triggered.connect(self._rebuild_python_legacy)

        ui_menu['menu.tools.system.rebuild'] = QAction(reload_icon, "Rebuild System Sandbox Docker Image", window)
        menu_tools.addAction(ui_menu['menu.tools.system.rebuild'])
        ui_menu['menu.tools.system.rebuild'].triggered.connect(self._rebuild_system)


# controller/files.py

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.03 00:00:00                  #
# ================================================== #

import datetime
import os
import shutil
from typing import Optional, Union
from shutil import copy2
from urllib.parse import unquote

from PySide6.QtWidgets import QFileDialog, QApplication

from pygpt_net.core.filesystem.opener import Opener
from pygpt_net.utils import trans


class Files:
    def __init__(self, window=None):
        """
        Files controller

        :param window: Window instance
        """
        self.window = window
        self.uploaded_ids = []

    def selection_change(self):
        """Select on list change"""
        # TODO: implement this
        pass

    def delete_recursive(
            self,
            path: str,
            force: bool = False
    ):
        """
        Delete directory with all files

        :param path: path to directory
        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='files.delete.recursive',
                id=path,
                msg=trans('files.delete.recursive.confirm'),
            )
            return
        try:
            shutil.rmtree(path)  # delete directory with all files
            self.window.update_status(
                f"[OK] Deleted directory: {os.path.basename(path)}"
            )
            self.update_explorer()
        except Exception as e:
            self.window.core.debug.log(e)
            print(f"Error deleting directory: {path} - {e}")

    def touch_file(
            self,
            path: str,
            name: Optional[str] = None,
            force: bool = False
    ):
        """
        Touch empty file

        :param path: path to file
        :param name: filename
        :param force: force touch
        """
        if not force:
            self.window.ui.dialog['create'].id = 'touch'
            self.window.ui.dialog['create'].input.setText("")
            self.window.ui.dialog['create'].current = path
            self.window.ui.dialog['create'].show()
            self.window.ui.dialog['create'].input.setFocus()
            return

        self.window.ui.dialog['create'].close()
        try:
            filepath = os.path.join(path, name)
            open(filepath, 'a').close()
            self.window.update_status(
                f"[OK] Created file: {filepath}"
            )
            self.update_explorer()
        except Exception as e:
            self.window.core.debug.log(e)
            print(f"Error creating file: {path} - {e}")

    def delete(
            self,
            path: Union[str, list],
            force: bool = False
    ):
        """
        Delete file or directory

        :param path: path to file
        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='files.delete',
                id=path,
                msg=trans('files.delete.confirm'),
            )
            return

        if isinstance(path, list):
            for p in path:
                self.delete(p, True)
            return

        if os.path.isdir(path):
            # check if directory is not empty
            if os.listdir(path):
                self.delete_recursive(path)
                return
            else:
                self.delete_recursive(path, True)
        else:
            try:
                os.remove(path)
                self.window.update_status(
                   f"[OK] Deleted file: {os.path.basename(path)}"
                )
                self.update_explorer()
            except Exception as e:
                self.window.core.debug.log(e)
                print(f"Error deleting file: {path} - {e}")

    def duplicate_local(
            self,
            path: Union[str, list],
            name: str,
            force: bool = False
    ):
        """
        Duplicate file or directory

        :param path: path to file or list of files
        :param name: new file name
        :param force: force duplicate
        """
        if isinstance(path, list):
            for p in path:
                if os.path.isdir(p):
                    new_name = os.path.basename(p) + "_copy"
                else:
                    new_name = os.path.splitext(os.path.basename(p))[0] \
                               + "_copy" + os.path.splitext(p)[1]
                self.duplicate_local(p, new_name, True)
            return

        if not force:
            if os.path.isdir(path):
                new_name = os.path.basename(path) + "_copy"
            else:
                new_name = os.path.splitext(os.path.basename(path))[0] \
                           + "_copy" + os.path.splitext(path)[1]
            self.window.ui.dialog['create'].id = 'duplicate'
            self.window.ui.dialog['create'].input.setText(new_name)
            self.window.ui.dialog['create'].current = path
            self.window.ui.dialog['create'].show()
            self.window.ui.dialog['create'].input.setFocus()
            return

        self.window.ui.dialog['create'].close()
        self.update_explorer()

        try:
            parent_dir = os.path.dirname(path)
            new_path = os.path.join(parent_dir, name)

            if os.path.exists(new_path):
                # prefix with timestamp if file exists
                ts_prefix = self.make_ts_prefix()
                name = f"{ts_prefix}_{name}"
                new_path = os.path.join(parent_dir, name)

            if os.path.isdir(path):
                shutil.copytree(path, new_path)
            else:
                copy2(path, new_path)
            self.window.update_status(
                f"[OK] Duplicated file: {os.path.basename(path)} -> {name}"
            )
            self.update_explorer()
        except Exception as e:
            self.window.core.debug.log(e)
            print(f"Error duplicating file: {path} - {e}")

    def download_local(self, path: Union[str, list]):
        """
        Download (copy) file or directory to local filesystem.

        Behavior:
        - Single selection: unchanged, prompts Save As dialog per current implementation.
        - Multi selection (>1 item): prompts once for a target directory and copies all selected items there.
          In case of name collision in the target directory, a timestamp prefix is added to the copied name.
        """
        # Multi-selection: choose a directory once and copy all into it
        if isinstance(path, list):
            # Normalize incoming list (decode, map to workdir)
            norm_paths = []
            for p in path:
                try:
                    p_norm = self.window.core.filesystem.to_workdir(unquote(p))
                except Exception:
                    p_norm = p
                if p_norm:
                    norm_paths.append(p_norm)

            if not norm_paths:
                return

            if len(norm_paths) == 1:
                # Defer to single-item path handling
                self.download_local(norm_paths[0])
                return

            last_dir = self.window.core.config.get_last_used_dir()
            target_dir = QFileDialog.getExistingDirectory(
                self.window,
                "Select target directory",
                last_dir if last_dir else os.path.expanduser("~"),
            )
            if not target_dir:
                return

            # Remember last used directory
            self.window.core.config.set_last_used_dir(target_dir)

            copied = 0

            for src in norm_paths:
                try:
                    if not os.path.exists(src):
                        continue

                    base_name = os.path.basename(src.rstrip(os.sep))
                    dst = os.path.join(target_dir, base_name)

                    # Avoid copying into itself and handle name collisions
                    if os.path.abspath(dst) == os.path.abspath(src) or os.path.exists(dst):
                        dst = os.path.join(target_dir, f"{self.make_ts_prefix()}_{base_name}")

                    if os.path.isdir(src):
                        shutil.copytree(src, dst)
                    else:
                        copy2(src, dst)
                    copied += 1
                except Exception as e:
                    self.window.core.debug.log(e)
                    print(f"Error downloading item: {src} -> {target_dir} - {e}")

            if copied > 0:
                self.window.update_status(f"[OK] Downloaded: {copied} items to: {target_dir}")
            return

        # Single-item flow (unchanged)
        path = self.window.core.filesystem.to_workdir(unquote(path))
        last_dir = self.window.core.config.get_last_used_dir()
        dialog = QFileDialog(self.window)
        dialog.setDirectory(last_dir)
        dialog.selectFile(os.path.basename(path))
        dialog.setAcceptMode(QFileDialog.AcceptSave)
        if dialog.exec():
            files = dialog.selectedFiles()
            if files:
                self.window.core.config.set_last_used_dir(
                    os.path.dirname(files[0])
                )
                try:
                    if os.path.isdir(path):
                        shutil.copytree(path, files[0])
                    else:
                        shutil.copy2(path, files[0])
                    self.window.update_status(
                        f"[OK] Downloaded file: {os.path.basename(path)}"
                    )
                except Exception as e:
                    self.window.core.debug.log(e)
                    print(f"Error downloading file: {path} - {e}")

    def upload_local(
            self,
            parent_path: Optional[str] = None
    ):
        """
        Upload local file(s) to directory

        :param parent_path: parent path
        """
        last_dir = self.window.core.config.get_last_used_dir()
        dialog = QFileDialog(self.window)
        dialog.setDirectory(last_dir)
        dialog.setFileMode(QFileDialog.ExistingFiles)
        if dialog.exec():
            files = dialog.selectedFiles()
            if files:
                self.window.core.config.set_last_used_dir(
                    os.path.dirname(files[0])
                )
                if parent_path:
                    target_directory = parent_path
                else:
                    target_directory = self.window.core.config.get_user_dir('data')
                num = 0
                for file_path in files:
                    path_to = os.path.join(
                        target_directory,
                        os.path.basename(file_path),
                    )
                    try:
                        # if exists, append timestamp
                        if os.path.exists(path_to):
                            new_name = self.make_ts_prefix() + "_" + os.path.basename(file_path)
                            target_path = os.path.join(
                                target_directory,
                                new_name,
                            )
                        else:
                            target_path = os.path.join(
                                target_directory,
                                os.path.basename(file_path),
                            )
                        if not os.path.exists(target_directory):
                            os.makedirs(
                                target_directory,
                                exist_ok=True,
                            )
                        copy2(file_path, target_path)
                        num += 1
                    except Exception as e:
                        self.window.core.debug.log(e)
                        print(f"Error copying file {file_path}: {e}")
                if num > 0:
                    self.window.update_status(f"[OK] Uploaded: {num} files.")
                    self.update_explorer()

    def upload_paths(
            self,
            paths: list,
            target_directory: Optional[str] = None
    ):
        """
        Upload provided local paths (files or directories) into target directory.
        - Directories are copied recursively.
        - Name collisions are resolved using timestamp prefix, consistent with upload_local().
        - Skips copying directory into itself or its subdirectory.

        :param paths: list of absolute local paths
        :param target_directory: destination directory (defaults to user 'data' dir)
        """
        if not paths:
            return
        if target_directory is None:
            target_directory = self.window.core.config.get_user_dir('data')

        try:
            if not os.path.exists(target_directory):
                os.makedirs(target_directory, exist_ok=True)
        except Exception as e:
            self.window.core.debug.log(e)
            return

        copied = 0

        def unique_dest(dest_path: str) -> str:
            if not os.path.exists(dest_path):
                return dest_path
            base_dir = os.path.dirname(dest_path)
            name = os.path.basename(dest_path)
            new_name = self.make_ts_prefix() + "_" + name
            return os.path.join(base_dir, new_name)

        for src in paths:
            try:
                if not src or not os.path.exists(src):
                    continue

                # Prevent copying a directory into itself or its subdirectory
                try:
                    if os.path.isdir(src):
                        common = os.path.commonpath([os.path.abspath(src), os.path.abspath(target_directory)])
                        if common == os.path.abspath(src):
                            # target is inside src; skip
                            self.window.core.debug.log(f"Skipped copying directory into itself: {src} -> {target_directory}")
                            continue
                except Exception:
                    pass

                dest_base = os.path.join(target_directory, os.path.basename(src))
                dest_path = unique_dest(dest_base)

                if os.path.isdir(src):
                    shutil.copytree(src, dest_path)
                    copied += 1
                else:
                    copy2(src, dest_path)
                    copied += 1
            except Exception as e:
                self.window.core.debug.log(e)
                print(f"Error uploading path {src}: {e}")

        if copied > 0:
            self.window.update_status(f"[OK] Uploaded: {copied} files.")
            self.update_explorer()

    def rename(self, path: Union[str, list]):
        """
        Rename file or directory

        :param path: path to file or list of files
        """
        if isinstance(path, list):
            for p in path:
                self.rename(p)
                return
        self.window.ui.dialog['rename'].id = 'output_file'
        self.window.ui.dialog['rename'].input.setText(os.path.basename(path))
        self.window.ui.dialog['rename'].current = path
        self.window.ui.dialog['rename'].show()
        self.window.ui.dialog['rename'].input.setFocus()

    def update_name(
            self,
            path: str,
            name: str
    ):
        """
        Update name of file or directory

        :param path: path
        :param name: name
        """
        new_path = os.path.join(os.path.dirname(path), name)
        if os.path.exists(new_path):
            self.window.update_status(
                f"[ERROR] File already exists: {os.path.basename(new_path)}"
            )
            return
        os.rename(path, new_path)
        self.window.update_status(
           f"[OK] Renamed: {os.path.basename(path)} -> {name}"
        )
        self.window.ui.dialog['rename'].close()
        self.update_explorer()

    def open_dir(
            self,
            path: Union[str, list],
            select: bool = False
    ):
        """
        Open file or directory in file manager

        :param path: path to file or directory or list of paths
        :param select: select file in file manager
        """
        if isinstance(path, list):
            parents = []
            for p in path:
                parent = os.path.dirname(p)
                if parent not in parents:
                    self.open_in_file_manager(p, select)
                if parent not in parents:
                    parents.append(parent)
            return
        self.open_in_file_manager(path, select)

    def open(self, path: Union[str, list]):
        """
        Open path in file manager or with default application

        :param path: path to file or directory or list of paths
        """
        if isinstance(path, list):
            for p in path:
                self.open(p)
            return
        path = self.window.core.filesystem.get_path(path)
        Opener.open_path(path, reveal=False)

    def open_in_file_manager(
            self,
            path: str,
            select: bool = False
    ):
        """
        Open in file manager

        :param path: path to file or directory
        :param select: select file in file manager
        """
        path = self.window.core.filesystem.get_path(path)
        if os.path.exists(path):
            Opener.open_path(path, reveal=select)

    def make_dir_dialog(self, path: str):
        """
        Open make dir directory name dialog

        :param path: path to file or directory
        """
        self.window.ui.dialog['create'].id = 'mkdir'
        self.window.ui.dialog['create'].input.setText("")
        self.window.ui.dialog['create'].current = path
        self.window.ui.dialog['create'].show()
        self.window.ui.dialog['create'].input.setFocus()

    def make_dir(
            self,
            path: str,
            name: Optional[str] = None
    ):
        """
        Make directory

        :param path: path to directory
        :param name: name of directory
        """
        self.window.ui.dialog['create'].close()
        if name is None:
            self.window.update_status(
                "[ERROR] Directory name is empty."
            )
            return
        path_dir = os.path.join(path, name)
        if os.path.exists(path_dir):
            self.window.update_status(
                "[ERROR] Directory or file already exists."
            )
            return
        os.makedirs(path_dir, exist_ok=True)
        self.window.update_status(
            f"[OK] Directory created: {name}"
        )
        self.update_explorer()

    def update_explorer(self, reload: bool = False):
        """
        Update file explorer view

        :param: reload: reload explorer root dir
        """
        data = self.window.core.idx.get_idx_data()  # get all files data, from all indexes
        self.window.ui.nodes['output_files'].index_data = data
        if reload:
            root = self.window.core.config.get_user_dir('data')
            self.window.ui.nodes['output_files'].directory = root
            self.window.ui.nodes['output_files'].update_view()
            self.window.ui.nodes['output_files'].path_label.setText(root)
        self.window.ui.nodes['output_files'].model.update_idx_status(data)

    def use_attachment(self, path: Union[str, list]):
        """
        Use file as attachment

        :param path: path to file or list of files
        """
        paths = path if isinstance(path, list) else [path]
        for p in paths:
            title = os.path.basename(p)
            mode = self.window.core.config.get("mode")
            self.window.core.attachments.new(
                mode=mode,
                name=title,
                path=p,
                auto_save=False,
            )
        self.window.core.attachments.save()
        self.window.controller.attachment.update()

    def copy_sys_path(self, path: Union[str, list]):
        """
        Copy system path to clipboard

        :param path: path to file or list of files
        """
        if isinstance(path, list):
            paths = [self.window.core.filesystem.get_path(p) for p in path]
            path_str = "\n".join(paths)
        else:
            path_str = self.window.core.filesystem.get_path(path)
        QApplication.clipboard().setText(path_str)
        self.window.controller.chat.common.append_to_input(path_str)

    def copy_work_path(self, path: Union[str, list]):
        """
        Copy work path to clipboard

        :param path: path to file  or list of files
        """
        if isinstance(path, list):
            paths = [self.strip_work_path(p) for p in path]
            path = "\n".join(paths)
        else:
            path = self.strip_work_path(path)
        QApplication.clipboard().setText(path)
        self.window.controller.chat.common.append_to_input(path)

    def make_read_cmd(self, path: Union[str, list]):
        """
        Make read command for file or directory and append to input

        :param path: path to file or list of files
        """
        files_list = path if isinstance(path, list) else [path]
        cmd = ""
        cmd_dir = []
        cmd_current = []
        for path in files_list:
            if os.path.isdir(path):
                cmd_dir.append(self.strip_work_path(path))
            else:
                cmd_current.append(self.strip_work_path(path))
        if len(cmd_dir) > 1:
            cmd = "Please list files from directories: " + ", ".join(cmd_dir)
        elif len(cmd_dir) == 1:
            cmd = f"Please list files from directory: {cmd_dir[0]}"
        if len(cmd_current) > 1:
            cmd = "Please read these files from current directory: " + ", ".join(cmd_current)
        elif len(cmd_current) == 1:
            cmd = f"Please read this file from current directory: {cmd_current[0]}"
        self.window.controller.chat.common.append_to_input(cmd)

    def make_ts_prefix(self) -> str:
        """
        Make timestamp prefix
        """
        return datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

    def strip_work_path(self, path: str) -> str:
        """
        Strip work path

        :param path: path to file
        :return: stripped path
        """
        work_dir = self.window.core.config.get_user_dir("data")
        path = path.replace(work_dir, "")
        if path.startswith("/") or path.startswith("\\"):
            path = path[1:]
        return path

    def reset(self):
        """Reset uploaded IDs"""
        self.uploaded_ids = []

    def get_ids(self) -> list:
        """
        Get uploaded files IDs

        :return: uploaded files IDs
        """
        return self.uploaded_ids

    def reload(self):
        """Reload files"""
        self.update_explorer(reload=True)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.17 19:00:00                  #
# ================================================== #

import json
import time
from typing import Optional, Dict, Any, List

from pygpt_net.core.types import (
    MODE_CHAT,
    MODE_VISION,
    MODE_AUDIO,
    MULTIMODAL_IMAGE,
    MODE_RESEARCH,
    MODE_COMPLETION,
    OPENAI_DISABLE_TOOLS,
)
from pygpt_net.core.bridge.context import BridgeContext, MultimodalContext
from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.model import ModelItem

from .utils import sanitize_name
from pygpt_net.item.attachment import AttachmentItem


class Chat:
    def __init__(self, window=None):
        """
        Chat wrapper

        :param window: Window instance
        """
        self.window = window
        self.input_tokens = 0
        self.audio_prev_id = None
        self.audio_prev_expires_ts = None

    def send(
            self,
            context: BridgeContext,
            extra: Optional[Dict[str, Any]] = None
    ):
        """
        Call OpenAI API for chat

        :param context: Bridge context
        :param extra: Extra arguments
        :return: response or stream chunks
        """
        prompt = context.prompt
        stream = context.stream
        max_tokens = int(context.max_tokens or 0)
        system_prompt = context.system_prompt
        mode = context.mode
        model = context.model
        functions = context.external_functions
        attachments = context.attachments
        multimodal_ctx = context.multimodal_ctx

        ctx = context.ctx
        if ctx is None:
            ctx = CtxItem()  # create empty context
        user_name = ctx.input_name  # from ctx
        ai_name = ctx.output_name  # from ctx

        client = self.window.core.api.openai.get_client(mode, context.model)

        # build chat messages
        messages = self.build(
            prompt=prompt,
            system_prompt=system_prompt,
            model=model,
            history=context.history,
            attachments=attachments,
            ai_name=ai_name,
            user_name=user_name,
            multimodal_ctx=multimodal_ctx,
        )
        msg_tokens = self.window.core.tokens.from_messages(
            messages,
            model.id,
        )
        # check if max tokens not exceeded
        if max_tokens > 0 and int(model.ctx) > 0:
            if msg_tokens + int(max_tokens) > model.ctx:
                max_tokens = model.ctx - msg_tokens - 1
                if max_tokens < 0:
                    max_tokens = 0

        # extra API kwargs
        response_kwargs = {}

        # tools / functions
        tools = self.window.core.api.openai.tools.prepare(model, functions)

        # fix: o1 compatibility
        if (model.id is not None
                and not model.id.startswith("o1")
                and not model.id.startswith("o3")
                and not model.id.startswith("o4")
                and model.is_gpt()):
            response_kwargs['presence_penalty'] = self.window.core.config.get('presence_penalty')
            response_kwargs['frequency_penalty'] = self.window.core.config.get('frequency_penalty')
            response_kwargs['temperature'] = self.window.core.config.get('temperature')
            response_kwargs['top_p'] = self.window.core.config.get('top_p')

        # extra arguments, o3 only
        if model.extra and "reasoning_effort" in model.extra:
            response_kwargs['reasoning_effort'] = model.extra["reasoning_effort"]

        # tool calls are not supported for some models
        if model.id in OPENAI_DISABLE_TOOLS:
            tools = []  # force disable

        # append tools
        if len(tools) > 0:
            response_kwargs['tools'] = tools

        if max_tokens > 0:
            if model.id is None or (not model.id.startswith("o1") and not model.id.startswith("o3")):
                response_kwargs['max_tokens'] = max_tokens
            else:
                response_kwargs['max_completion_tokens'] = max_tokens

        # audio mode
        if mode in [MODE_AUDIO]:
            stream = False
            voice_id = "alloy"
            tmp_voice = self.window.core.plugins.get_option("audio_output", "openai_voice")
            if tmp_voice:
                voice_id = tmp_voice
            response_kwargs["modalities"] = ["text", "audio"]
            response_kwargs["audio"] = {
                "voice": voice_id,
                "format": "wav"
            }

        if stream:
            response_kwargs['stream_options'] = {"include_usage": True}

        # OpenRouter: add web search remote tool (if enabled)
        model_id = model.id
        if model.provider == "open_router":
            model_id = self.window.core.models.get_openrouter_model(model)

        response = client.chat.completions.create(
            messages=messages,
            model=model_id,
            stream=stream,
            **response_kwargs,
        )
        return response

    def build(
            self,
            prompt: str,
            system_prompt: str,
            model: ModelItem,
            history: Optional[List[CtxItem]] = None,
            attachments: Optional[Dict[str, AttachmentItem]] = None,
            ai_name: Optional[str] = None,
            user_name: Optional[str] = None,
            multimodal_ctx: Optional[MultimodalContext] = None,
    ) -> list:
        """
        Build list of chat messages

        :param prompt: user prompt
        :param system_prompt: system prompt
        :param history: history
        :param model: model item
        :param attachments: attachments
        :param ai_name: AI name
        :param user_name: username
        :param multimodal_ctx: Multimodal context
        :return: messages list
        """
        self.window.core.api.openai.vision.reset()
        messages = []

        # tokens config
        mode = MODE_CHAT
        is_tool_output = False
        tool_call_native_enabled = self.window.core.config.get('func_call.native', False)
        allowed_system = True
        if (model.id is not None
                and model.id in ["o1-mini", "o1-preview"]):
            allowed_system = False

        used_tokens = self.window.core.tokens.from_user(
            prompt,
            system_prompt,
        )  # threshold and extra included
        max_ctx_tokens = self.window.core.config.get('max_total_tokens')  # max context window

        # fit to max model tokens
        if max_ctx_tokens > model.ctx > 0:
            max_ctx_tokens = model.ctx

        # input tokens: reset
        self.reset_tokens()

        # names fallback
        if ai_name is None or ai_name == "":
            ai_name = "assistant"
        if user_name is None or user_name == "":
            user_name = "user"

        # append system prompt
        if allowed_system:
            if system_prompt is not None and system_prompt != "":
                messages.append({"role": "system", "content": system_prompt})

        # append messages from context (memory)
        if self.window.core.config.get('use_context'):
            items = self.window.core.ctx.get_history(
                history,
                model.id,
                mode,
                used_tokens,
                max_ctx_tokens,
            )
            for item in items:
                # input
                if item.final_input is not None and item.final_input != "":
                    messages.append({
                        "role": "user",
                        "name": sanitize_name(user_name),
                        "content": item.final_input,
                    })

                # output
                if item.final_output is not None and item.final_output != "":
                    msg = {
                        "role": "assistant",
                        "name": sanitize_name(ai_name),
                        "content": item.final_output,
                    }
                    # append previous audio ID
                    if MODE_AUDIO in model.mode:
                        if item.audio_id:
                            # at first check expires_at - expired audio throws error in API
                            current_timestamp = time.time()
                            audio_timestamp = int(item.audio_expires_ts) if item.audio_expires_ts else 0
                            if audio_timestamp and audio_timestamp > current_timestamp:
                                msg["audio"] = {
                                    "id": item.audio_id
                                }
                        elif self.audio_prev_id:
                            current_timestamp = time.time()
                            audio_timestamp = int(self.audio_prev_expires_ts) if self.audio_prev_expires_ts else 0
                            if audio_timestamp and audio_timestamp > current_timestamp:
                                msg["audio"] = {
                                    "id": self.audio_prev_id
                                }
                    messages.append(msg)

                    # ---- tool output ----
                    is_tool_output = False
                    is_last_item = item == items[-1] if items else False
                    if is_last_item and tool_call_native_enabled and item.extra and isinstance(item.extra, dict):
                        if "tool_calls" in item.extra and isinstance(item.extra["tool_calls"], list):
                            for tool_call in item.extra["tool_calls"]:
                                if "function" in tool_call:
                                    if "id" not in tool_call or "name" not in tool_call["function"]:
                                        continue
                                    if tool_call["id"] and tool_call["function"]["name"]:
                                        if "tool_output" in item.extra and isinstance(item.extra["tool_output"], list):
                                            for tool_output in item.extra["tool_output"]:
                                                if ("cmd" in tool_output
                                                        and tool_output["cmd"] == tool_call["function"]["name"]):
                                                    msg = {
                                                        "role": "tool",
                                                        "tool_call_id": tool_call["id"],
                                                        "content": str(tool_output),
                                                    }
                                                    last_msg = messages[-1] if messages else None
                                                    if last_msg and last_msg.get(
                                                            "role") == "assistant":
                                                        last_msg["tool_calls"] = []
                                                        for call in item.extra["tool_calls"]:
                                                            last_msg["tool_calls"].append(
                                                                {
                                                                    "id": call["id"],
                                                                    "type": "function",
                                                                    "function": {
                                                                        "name": call["function"]["name"],
                                                                        "arguments": json.dumps(
                                                                            call["function"]["arguments"]),
                                                                    }
                                                                }
                                                            )
                                                        last_msg["content"] = ""
                                                    messages.append(msg)
                                                    is_tool_output = True
                                                    break
                                                elif "result" in tool_output:
                                                    # if result is present, append it as function call output
                                                    msg = {
                                                        "role": "tool",
                                                        "tool_call_id": tool_call["id"],
                                                        "content": str(tool_output["result"]),
                                                    }
                                                    last_msg = messages[-1] if messages else None
                                                    if last_msg and last_msg.get(
                                                            "role") == "assistant":
                                                        last_msg["tool_calls"] = []
                                                        for call in item.extra["tool_calls"]:
                                                            last_msg["tool_calls"].append(
                                                                {
                                                                    "id": call["id"],
                                                                    "type": "function",
                                                                    "function": {
                                                                        "name": call["function"]["name"],
                                                                        "arguments": json.dumps(call["function"]["arguments"]),
                                                                    }
                                                                }
                                                            )
                                                        last_msg["content"] = ""
                                                    messages.append(msg)
                                                    is_tool_output = True
                                                    break

        # use vision and audio if available in current model
        if not is_tool_output:  # append current prompt only if not tool output
            content = str(prompt)
            if model.is_image_input():
                content = self.window.core.api.openai.vision.build_content(
                    content=content,
                    attachments=attachments,
                )
            if model.is_audio_input():
                content = self.window.core.api.openai.audio.build_content(
                    content=content,
                    multimodal_ctx=multimodal_ctx,
                )

            # append current prompt
            messages.append({
                "role": "user",
                "content": content,
            })

        # input tokens: update
        self.input_tokens += self.window.core.tokens.from_messages(
            messages,
            model.id,
        )
        return messages

    def reset_tokens(self):
        """Reset input tokens counter"""
        self.input_tokens = 0

    def get_used_tokens(self) -> int:
        """
        Get input tokens counter

        :return: input tokens
        """
        return self.input_tokens

    def unpack_response(self, mode: str, response, ctx: CtxItem):
        """
        Unpack response from OpenAI API and set context

        :param mode: str - mode of the response (chat, vision, audio)
        :param response: OpenAI API response object
        :param ctx: CtxItem - context item to set the response data
        """
        output = ""
        if mode == MODE_COMPLETION:
            output = response.choices[0].text.strip()
        elif mode in [
            MODE_CHAT,
            # MODE_VISION,
            MODE_RESEARCH,
        ]:
            if response.choices[0]:
                if response.choices[0].message.content:
                    output = response.choices[0].message.content.strip()
                if response.choices[0].message.tool_calls:
                    ctx.tool_calls = self.window.core.command.unpack_tool_calls(
                        response.choices[0].message.tool_calls,
                    )
        # audio
        elif mode in [MODE_AUDIO]:
            if response.choices[0]:
                if response.choices[0].message and response.choices[0].message.audio:
                    ctx.audio_output = response.choices[0].message.audio.data
                    ctx.audio_id = response.choices[0].message.audio.id
                    ctx.audio_expires_ts = response.choices[0].message.audio.expires_at
                    ctx.is_audio = True
                    output = response.choices[0].message.audio.transcript  # from transcript
                    self.audio_prev_expires_ts = ctx.audio_expires_ts
                    self.audio_prev_id = ctx.audio_id
                elif response.choices[0].message and not response.choices[0].message.audio:
                    output = response.choices[0].message.content
                    ctx.audio_id = self.audio_prev_id
                    ctx.audio_expires_ts = self.audio_prev_expires_ts
                if response.choices[0].message.tool_calls:
                    ctx.tool_calls = self.window.core.command.unpack_tool_calls(
                        response.choices[0].message.tool_calls,
                    )

        ctx.output = output  # set output text
        ctx.set_tokens(
            response.usage.prompt_tokens,
            response.usage.completion_tokens,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.02 20:00:00                  #
# ================================================== #

from .config import Config
from .core.access import Access
from .core.agents import Agents
from .core.assistants import Assistants
from .core.attachments import Attachments
from .core.audio import Audio
from .core.bridge import Bridge
from .core.calendar import Calendar
from .core.camera import Camera
# from .core.chain import Chain
from .core.command import Command
from .core.ctx import Ctx
from .core.db import Database
from .core.debug import Debug
from .core.dispatcher import Dispatcher
from .core.experts import Experts
from .core.idx import Idx
from .core.installer import Installer
from .core.filesystem import Filesystem
from .core.history import History
from .core.image import Image
from .core.llm import LLM
from .core.models import Models
from .core.modes import Modes
from .core.notepad import Notepad
from .core.platforms import Platforms
from .core.plugins import Plugins
from .core.presets import Presets
from .core.prompt import Prompt
from .core.remote_store import RemoteStore
from .core.settings import Settings
from .core.tabs import Tabs
from .core.text import Text
from .core.tokens import Tokens
from .core.updater import Updater
from .core.video import Video
from .core.vision import Vision
from .core.web import Web

from .provider.api import Api

class Core:
    def __init__(self, window=None):
        """
        App core

        :param window: Window instance
        """
        self.window = window

        self.access = Access(window)
        self.agents = Agents(window)
        self.api = Api(window)
        self.assistants = Assistants(window)
        self.attachments = Attachments(window)
        self.audio = Audio(window)
        self.bridge = Bridge(window)
        self.calendar = Calendar(window)
        self.camera = Camera(window)
        # self.chain = Chain(window)  # deprecated from v2.5.20
        self.command = Command(window)
        self.config = Config(window)
        self.ctx = Ctx(window)
        self.db = Database(window)
        self.debug = Debug(window)
        self.dispatcher = Dispatcher(window)
        self.experts = Experts(window)
        self.filesystem = Filesystem(window)
        self.history = History(window)
        self.idx = Idx(window)
        self.image = Image(window)
        self.llm = LLM(window)
        self.installer = Installer(window)
        self.models = Models(window)
        self.modes = Modes(window)
        self.notepad = Notepad(window)
        self.platforms = Platforms(window)
        self.plugins = Plugins(window)
        self.presets = Presets(window)
        self.prompt = Prompt(window)        
        self.remote_store = RemoteStore(window)
        self.settings = Settings(window)
        self.tabs = Tabs(window)
        self.text = Text(window)
        self.tokens = Tokens(window)
        self.updater = Updater(window)
        self.video = Video(window)
        self.vision = Vision(window)
        self.web = Web(window)

    def init(self):
        """Initialize all components"""
        self.config.init(all=True)
        self.platforms.init()

    def patch(self):
        """Patch version"""
        self.updater.patch()

    def post_setup(self):
        """Post setup"""
        self.config.setup_env()

    def reload(self):
        """Reload core"""
        self.db.reload()
        self.patch()
        self.debug.update_logger_path()
        self.config.setup_env()
        self.prompt.custom.reload()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.14 00:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        # Endpoints / HTTP
        plugin.add_option(
            "graph_version",
            type="text",
            value="v21.0",
            label="Graph API version",
            description="API version (e.g. v21.0).",
        )
        plugin.add_option(
            "api_base",
            type="text",
            value="https://graph.facebook.com",
            label="API base",
            description="Base API host (Graph API). Version is appended automatically.",
        )
        plugin.add_option(
            "authorize_base",
            type="text",
            value="https://www.facebook.com",
            label="Authorize base",
            description="Base for OAuth authorize (Login dialog). Version is appended automatically.",
        )
        plugin.add_option(
            "http_timeout",
            type="int",
            value=30,
            label="HTTP timeout (s)",
            description="Requests timeout in seconds.",
        )

        # OAuth2 (PKCE)
        plugin.add_option(
            "oauth2_client_id",
            type="text",
            value="",
            label="App ID (client_id)",
            description="Facebook App ID.",
            secret=False,
        )
        plugin.add_option(
            "oauth2_client_secret",
            type="text",
            value="",
            label="App Secret (optional)",
            description="Needed for long-lived token exchange and if not using PKCE.",
            secret=True,
        )
        plugin.add_option(
            "oauth2_confidential",
            type="bool",
            value=False,
            label="Confidential client (use client_secret on exchange)",
            description="If enabled, code exchange includes client_secret instead of code_verifier.",
        )
        plugin.add_option(
            "oauth2_redirect_uri",
            type="text",
            value="http://127.0.0.1:8732/callback",
            label="Redirect URI",
            description="Must exactly match one of the Valid OAuth Redirect URIs in your Meta App.",
        )
        plugin.add_option(
            "oauth2_scopes",
            type="text",
            value="public_profile pages_show_list pages_read_engagement pages_manage_posts pages_read_user_content openid",
            label="Scopes",
            description="Space-separated permissions for Authorization Code (PKCE).",
        )
        plugin.add_option(
            "oauth2_code_verifier",
            type="text",
            value="",
            label="(auto) code_verifier",
            description="Generated by fb_oauth_begin.",
            secret=True,
        )
        plugin.add_option(
            "oauth2_state",
            type="text",
            value="",
            label="(auto) state",
            description="Generated by fb_oauth_begin.",
            secret=True,
        )
        plugin.add_option(
            "oauth2_nonce",
            type="text",
            value="",
            label="(auto) nonce",
            description="Generated by fb_oauth_begin (OIDC).",
            secret=True,
        )
        plugin.add_option(
            "oauth2_access_token",
            type="textarea",
            value="",
            label="(auto) User access token",
            description="Stored user access token.",
            secret=True,
        )
        plugin.add_option(
            "oauth2_expires_at",
            type="text",
            value="0",
            label="(auto) Expires at (unix)",
            description="Auto-calculated expiry time.",
            secret=False,
        )

        # Convenience cache
        plugin.add_option(
            "user_id",
            type="text",
            value="",
            label="(auto) User ID",
            description="Cached after fb_me or oauth exchange.",
        )
        plugin.add_option(
            "user_name",
            type="text",
            value="",
            label="(auto) User name",
            description="Cached after fb_me or oauth exchange.",
        )
        plugin.add_option(
            "fb_page_id",
            type="text",
            value="",
            label="(auto) Default Page ID",
            description="Chosen by fb_page_set_default.",
        )
        plugin.add_option(
            "fb_page_name",
            type="text",
            value="",
            label="(auto) Default Page name",
            description="Chosen by fb_page_set_default.",
        )
        plugin.add_option(
            "fb_page_access_token",
            type="textarea",
            value="",
            label="(auto) Default Page access token",
            description="Cached via fb_page_set_default or on demand.",
            secret=True,
        )

        # OAuth UX
        plugin.add_option(
            "oauth_auto_begin",
            type="bool",
            value=True,
            label="Auto-start OAuth when required",
            description="If a command needs user token, begin PKCE flow automatically.",
        )
        plugin.add_option(
            "oauth_open_browser",
            type="bool",
            value=True,
            label="Open browser automatically",
            description="Open authorize URL in default browser.",
        )
        plugin.add_option(
            "oauth_local_server",
            type="bool",
            value=True,
            label="Use local server for OAuth",
            description="Start local HTTP server to capture redirect.",
        )
        plugin.add_option(
            "oauth_local_timeout",
            type="int",
            value=180,
            label="OAuth local timeout (s)",
            description="How long to wait for redirect with code.",
        )
        plugin.add_option(
            "oauth_success_html",
            type="textarea",
            value="<html><body><h3>Authorization complete. You can close this window.</h3></body></html>",
            label="Success HTML",
            description="HTML shown on local callback success.",
        )
        plugin.add_option(
            "oauth_fail_html",
            type="textarea",
            value="<html><body><h3>Authorization failed.</h3></body></html>",
            label="Fail HTML",
            description="HTML shown on local callback error.",
        )
        plugin.add_option(
            "oauth_local_port",
            type="int",
            value=8732,
            label="OAuth local port (0=auto)",
            description="Local HTTP port for callback; use >1024. Must be allowed in App.",
        )
        plugin.add_option(
            "oauth_allow_port_fallback",
            type="bool",
            value=True,
            label="Allow fallback port if busy",
            description="If preferred port is busy/forbidden, pick a free local port.",
        )

        # ---------------- Commands ----------------

        # Auth
        plugin.add_cmd(
            "fb_oauth_begin",
            instruction="Begin OAuth2 (PKCE) flow (returns authorize URL).",
            params=[
                {"name": "scopes", "type": "str", "required": False, "description": "Override permissions (space-separated)"},
                {"name": "state", "type": "str", "required": False, "description": "Optional CSRF state"},
            ],
            enabled=True,
            description="Auth: begin OAuth2",
            tab="auth",
        )
        plugin.add_cmd(
            "fb_oauth_exchange",
            instruction="Exchange authorization code for user access token.",
            params=[
                {"name": "code", "type": "str", "required": True, "description": "Authorization code"},
                {"name": "state", "type": "str", "required": False, "description": "State (if used)"},
            ],
            enabled=True,
            description="Auth: exchange code",
            tab="auth",
        )
        plugin.add_cmd(
            "fb_token_extend",
            instruction="Exchange short-lived user token for long-lived token (requires app secret).",
            params=[],
            enabled=True,
            description="Auth: extend user token",
            tab="auth",
        )

        # Me
        plugin.add_cmd(
            "fb_me",
            instruction="Get authorized user profile.",
            params=[
                {"name": "fields", "type": "str", "required": False, "description": "Default: id,name"},
            ],
            enabled=True,
            description="Users: me",
            tab="users",
        )

        # Pages
        plugin.add_cmd(
            "fb_pages_list",
            instruction="List Pages I manage (id,name,access_token,tasks).",
            params=[
                {"name": "fields", "type": "str", "required": False, "description": "Default: id,name,access_token,tasks"},
                {"name": "limit", "type": "int", "required": False, "description": "Default 25"},
                {"name": "after", "type": "str", "required": False, "description": "Paging cursor"},
                {"name": "before", "type": "str", "required": False, "description": "Paging cursor"},
            ],
            enabled=True,
            description="Pages: list",
            tab="pages",
        )
        plugin.add_cmd(
            "fb_page_set_default",
            instruction="Set default Page (caches name and access token).",
            params=[
                {"name": "page_id", "type": "str", "required": True, "description": "Page ID"},
                {"name": "fetch_token", "type": "bool", "required": False, "description": "Fetch and cache page access token"},
            ],
            enabled=True,
            description="Pages: set default",
            tab="pages",
        )

        # Posts
        plugin.add_cmd(
            "fb_page_posts",
            instruction="Get Page feed (posts).",
            params=[
                {"name": "page_id", "type": "str", "required": False, "description": "Defaults to selected page"},
                {"name": "fields", "type": "str", "required": False, "description": "Default: id,message,created_time,permalink_url,is_published"},
                {"name": "limit", "type": "int", "required": False, "description": "Default 25"},
                {"name": "since", "type": "str", "required": False, "description": "Unix or ISO8601"},
                {"name": "until", "type": "str", "required": False, "description": "Unix or ISO8601"},
                {"name": "after", "type": "str", "required": False, "description": "Paging cursor"},
                {"name": "before", "type": "str", "required": False, "description": "Paging cursor"},
            ],
            enabled=True,
            description="Posts: list",
            tab="posts",
        )
        plugin.add_cmd(
            "fb_page_post_create",
            instruction="Publish a Page post (text/link and optional photos).",
            params=[
                {"name": "page_id", "type": "str", "required": False, "description": "Defaults to selected page"},
                {"name": "message", "type": "str", "required": False, "description": "Post text"},
                {"name": "link", "type": "str", "required": False, "description": "URL link"},
                {"name": "published", "type": "bool", "required": False, "description": "Default true"},
                {"name": "scheduled_publish_time", "type": "str", "required": False, "description": "Unix seconds or ISO8601 (requires published=false)"},
                {"name": "media_fbids", "type": "list", "required": False, "description": "Photo IDs uploaded earlier"},
                {"name": "photo_urls", "type": "list", "required": False, "description": "Auto-upload URLs (unpublished then attach)"},
                {"name": "photo_paths", "type": "list", "required": False, "description": "Auto-upload local files (unpublished then attach)"},
                {"name": "targeting", "type": "dict", "required": False, "description": "Audience targeting object"},
            ],
            enabled=True,
            description="Posts: create",
            tab="posts",
        )
        plugin.add_cmd(
            "fb_page_post_delete",
            instruction="Delete a Page post.",
            params=[
                {"name": "post_id", "type": "str", "required": True, "description": "Page post ID (e.g. {pageid}_{postid})"},
            ],
            enabled=True,
            description="Posts: delete",
            tab="posts",
        )

        # Media
        plugin.add_cmd(
            "fb_page_photo_upload",
            instruction="Upload a photo to Page (local path or URL).",
            params=[
                {"name": "page_id", "type": "str", "required": False, "description": "Defaults to selected page"},
                {"name": "path", "type": "str", "required": False, "description": "Local file path"},
                {"name": "url", "type": "str", "required": False, "description": "Remote image URL"},
                {"name": "caption", "type": "str", "required": False, "description": "Photo caption"},
                {"name": "published", "type": "bool", "required": False, "description": "Default true"},
                {"name": "temporary", "type": "bool", "required": False, "description": "Required for scheduled multi-photo"},
            ],
            enabled=True,
            description="Media: upload photo",
            tab="media",
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.02 19:00:00                  #
# ================================================== #

from typing import Optional, List, Dict

from google.genai import types as gtypes
from llama_index.core.llms.llm import BaseLLM as LlamaBaseLLM
from llama_index.core.base.embeddings.base import BaseEmbedding

from pygpt_net.core.types import (
    MODE_LLAMA_INDEX,
)
from pygpt_net.provider.llms.base import BaseLLM
from pygpt_net.item.model import ModelItem


class GoogleLLM(BaseLLM):
    def __init__(self, *args, **kwargs):
        super(GoogleLLM, self).__init__(*args, **kwargs)
        """
        Required ENV variables:
            - GOOGLE_API_KEY - API key for Google API
        Required args:
            - model: model name, e.g. gemini-1,5-pro
            - api_key: API key for Google API
        """
        self.id = "google"
        self.name = "Google"
        self.type = [MODE_LLAMA_INDEX, "embeddings"]

    def llama(
            self,
            window,
            model: ModelItem,
            stream: bool = False
    ) -> LlamaBaseLLM:
        """
        Return LLM provider instance for llama

        :param window: window instance
        :param model: model instance
        :param stream: stream mode
        :return: LLM provider instance
        """
        from llama_index.llms.google_genai import GoogleGenAI
        args = self.parse_args(model.llama_index, window)
        if "model" not in args:
            args["model"] = model.id
        if "api_key" not in args or args["api_key"] == "":
            args["api_key"] = window.core.config.get("api_key_google", "")

        window.core.api.google.setup_env()  # setup VertexAI if configured
        args = self.inject_llamaindex_http_clients(args, window.core.config)

        # -----------------------------------------------------------
        # Remote built-in tools for Google GenAI via LlamaIndex:
        # - Google Search grounding (Tool(google_search=GoogleSearch()))
        # - Code Execution (Tool(code_execution=ToolCodeExecution()))
        # - Url Context (Tool(url_context=UrlContext)) on 2.x+
        # We reuse native builder and forward tools into LlamaIndex.
        # If 1 tool -> use 'built_in_tool', if >1 -> pack into generation_config.tools
        # -----------------------------------------------------------
        built_tools = []
        try:
            built_tools = window.core.api.google.remote_tools.build_remote_tools(model=model) or []
        except Exception as e:
            window.core.debug.log(e)

        if built_tools:
            # Only attach if user didn't already pass their own config
            if "built_in_tool" not in args and "generation_config" not in args:
                if len(built_tools) == 1:
                    args["built_in_tool"] = built_tools[0]
                else:
                    # If multiple tools are enabled, provide them via generation_config.tools
                    try:
                        args["generation_config"] = gtypes.GenerateContentConfig(tools=built_tools)
                    except Exception as e:
                        # Fallback to the first tool if GenerateContentConfig cannot be constructed
                        window.core.debug.log(e)
                        args["built_in_tool"] = built_tools[0]

        return GoogleGenAI(**args)

    def get_embeddings_model(
            self,
            window,
            config: Optional[List[Dict]] = None
    ) -> BaseEmbedding:
        """
        Return provider instance for embeddings

        :param window: window instance
        :param config: config keyword arguments list
        :return: Embedding provider instance
        """
        from llama_index.embeddings.google_genai import GoogleGenAIEmbedding
        args = {}
        if config is not None:
            args = self.parse_args({
                "args": config,
            }, window)
        if "api_key" not in args or args["api_key"] == "":
            args["api_key"] = window.core.config.get("api_key_google", "")
        if "model" in args and "model_name" not in args:
            args["model_name"] = args.pop("model")

        window.core.api.google.setup_env()  # setup VertexAI if configured
        args = self.inject_llamaindex_http_clients(args, window.core.config)
        return GoogleGenAIEmbedding(**args)

    def get_models(
            self,
            window,
    ) -> List[Dict]:
        """
        Return list of models for the provider

        :param window: window instance
        :return: list of models
        """
        items = []
        client = window.core.api.google.get_client()
        models_list = client.models.list()
        for item in models_list:
            id = item.name.replace("models/", "")
            items.append({
                "id": id,
                "name": id,  # TODO: token limit get from API
            })
        return items

    def inject_llamaindex_http_clients(self, args: dict, cfg) -> dict:
        proxy = cfg.get("api_proxy")
        if not cfg.get("api_proxy.enabled", False):
            proxy = ""
        if proxy:
            http_options = gtypes.HttpOptions(
                client_args={"proxy": proxy},
                async_client_args={"proxy": proxy},
            )
            args["http_options"] = http_options
        return args



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.24 23:00:00                  #
# ================================================== #

from __future__ import annotations
from typing import Protocol, Optional, Any


class Logger(Protocol):
    """Minimal logger protocol used across the flow core."""
    def debug(self, msg: str, **kwargs: Any) -> None: ...
    def info(self, msg: str, **kwargs: Any) -> None: ...
    def warning(self, msg: str, **kwargs: Any) -> None: ...
    def error(self, msg: str, **kwargs: Any) -> None: ...


class NullLogger:
    """No-op logger."""
    def debug(self, msg: str, **kwargs: Any) -> None: pass
    def info(self, msg: str, **kwargs: Any) -> None: pass
    def warning(self, msg: str, **kwargs: Any) -> None: pass
    def error(self, msg: str, **kwargs: Any) -> None: pass


class StdLogger:
    """Simple stdout logger; can be replaced by external one from your app."""
    def __init__(self, prefix: str = "[flow]"):
        self.prefix = prefix

    def _fmt(self, level: str, msg: str) -> str:
        return f"{self.prefix} {level.upper()}: {msg}"

    def debug(self, msg: str, **kwargs: Any) -> None:
        print(self._fmt("debug", msg))

    def info(self, msg: str, **kwargs: Any) -> None:
        print(self._fmt("info", msg))

    def warning(self, msg: str, **kwargs: Any) -> None:
        print(self._fmt("warning", msg))

    def error(self, msg: str, **kwargs: Any) -> None:
        print(self._fmt("error", msg))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.26 17:00:00                  #
# ================================================== #

import copy
from typing import Dict, Any, Tuple, Union, Optional

from agents import (
    Agent as OpenAIAgent,
    Runner,
    TResponseInputItem,
)

from pygpt_net.core.agents.bridge import ConnectionContext
from pygpt_net.core.bridge import BridgeContext
from pygpt_net.core.types import (
    AGENT_MODE_OPENAI,
    AGENT_TYPE_OPENAI,
)

from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.model import ModelItem
from pygpt_net.item.preset import PresetItem

from pygpt_net.provider.api.openai.agents.remote_tools import append_tools
from pygpt_net.provider.api.openai.agents.response import StreamHandler
from pygpt_net.provider.api.openai.agents.experts import get_experts
from pygpt_net.utils import trans

from ..base import BaseAgent


class Agent(BaseAgent):

    PROMPT_BOT_1 = (
        "You're an advanced AI assistant and an expert in every field. "
        "Imagine that I am also such an AI assistant and converse with me in an expert manner. "
        "As two AI assistants, let's brainstorm and arrive at some advanced solutions."
    )

    PROMPT_BOT_2 = (
        "You're an advanced AI assistant and an expert in every field. "
        "Imagine that I am also such an AI assistant and converse with me in an expert manner. "
        "As two AI assistants, let's brainstorm and arrive at some advanced solutions."
    )

    def __init__(self, *args, **kwargs):
        super(Agent, self).__init__(*args, **kwargs)
        self.id = "openai_agent_b2b"
        self.type = AGENT_TYPE_OPENAI
        self.mode = AGENT_MODE_OPENAI
        self.name = "Bot 2 Bot"

    def get_agent(self, window, kwargs: Dict[str, Any]):
        """
        Return Agent provider instance

        :param window: window instance
        :param kwargs: keyword arguments
        :return: Agent provider instance
        """
        context = kwargs.get("context", BridgeContext())
        preset = context.preset
        model = kwargs.get("model", ModelItem())
        tools = kwargs.get("function_tools", [])
        handoffs = kwargs.get("handoffs", [])
        id = kwargs.get("bot_id", 1)
        option_key = f"bot_{id}"
        kwargs = {
            "name": self.get_option(preset, option_key, "name"),
            "instructions": self.get_option(preset, option_key, "prompt"),
            "model": window.core.agents.provider.get_openai_model(model),
        }
        if handoffs:
            kwargs["handoffs"] = handoffs

        tool_kwargs = append_tools(
            tools=tools,
            window=window,
            model=model,
            preset=preset,
            allow_local_tools=self.get_option(preset, option_key, "allow_local_tools"),
            allow_remote_tools= self.get_option(preset, option_key, "allow_remote_tools"),
        )
        kwargs.update(tool_kwargs) # update kwargs with tools
        return OpenAIAgent(**kwargs)

    def reverse_history(
            self,
            items: list[TResponseInputItem]
    ):
        """
        Reverse the roles of items in the input list in history (assistant to user)

        :param items: List of input items
        :return: List of input items with reversed roles
        """
        counter = 1
        for item in items:
            if item.get("role") == "assistant":
                if counter % 2 == 0:
                    item["role"] = "user"
                counter += 1
        return items

    def reverse_items(
            self,
            items: list[TResponseInputItem],
            verbose: bool = False
    ) -> list[TResponseInputItem]:
        """
        Reverse input items for the second bot

        :param items: List of input items
        :param verbose: Whether to print debug information
        :return: Reversed list of input items
        """
        if verbose:
            print("--- BEFORE ---")
            for item in items:
                print(item)

        reversed_items = []
        for item in items:
            if item.get("role") == "assistant":
                if isinstance(item.get("content"), str):
                    content = item.get("content", "")
                    item = {
                        "role": "user",
                        "content": content,
                    }
                else:
                    content = item["content"][0].get("text", "")
                    item = {
                        "role": "user",
                        "content": content,
                    }
            elif item.get("role") == "user":
                if isinstance(item.get("content"), str):
                    content = item.get("content", "")
                else:
                    content = item["content"][0].get("text", "")
                item = {
                    "role": "assistant",
                    "content": [
                        {
                            "type": "output_text",
                            "text": content,
                        }
                    ],
                }
            reversed_items.append(item)

        if verbose:
            print("--- AFTER ---")
            for item in reversed_items:
                print(item)

        return reversed_items

    def prepare_model(
            self,
            model: ModelItem,
            window: Any,
            previous_response_id: Union[str, None],
            kwargs: Dict[str, Any]
    ):
        """
        Prepare model configuration for the agent

        :param model: ModelItem instance
        :param window: Window instance
        :param previous_response_id: ID of the previous response (if any)
        :param kwargs: Additional keyword arguments for the model configuration
        :return: Prepared keyword arguments for the model
        """
        if model.provider == "openai":
            if previous_response_id:
                kwargs["previous_response_id"] = previous_response_id
        return kwargs

    async def run(
            self,
            window: Any = None,
            agent_kwargs: Dict[str, Any] = None,
            previous_response_id: str = None,
            messages: list = None,
            ctx: CtxItem = None,
            stream: bool = False,
            bridge: ConnectionContext = None,
            use_partial_ctx: Optional[bool] = False,
    ) -> Tuple[CtxItem, str, str]:
        """
        Run agent (async)

        :param window: Window instance
        :param agent_kwargs: Additional agent parameters
        :param previous_response_id: ID of the previous response (if any)
        :param messages: Conversation messages
        :param ctx: Context item
        :param stream: Whether to stream output
        :param bridge: Connection context for handling stop and step events
        :param use_partial_ctx: Use partial ctx per cycle
        :return: Current ctx, final output, last response ID
        """
        final_output = ""
        response_id = None
        reverse_verbose = False
        model = agent_kwargs.get("model", ModelItem())
        verbose = agent_kwargs.get("verbose", False)
        max_steps = agent_kwargs.get("max_iterations", 10)
        preset = agent_kwargs.get("preset", PresetItem())
        tools = agent_kwargs.get("function_tools", [])

        # add experts
        experts = get_experts(
            window=window,
            preset=preset,
            verbose=verbose,
            tools=tools,
        )

        bot_1_name = self.get_option(preset, "bot_1", "name")
        bot_1_kwargs = copy.deepcopy(agent_kwargs)
        bot_2_kwargs = copy.deepcopy(agent_kwargs)

        bot_1_kwargs["bot_id"] = 1
        if experts:
            bot_1_kwargs["handoffs"] = experts
        bot_1 = self.get_agent(window, bot_1_kwargs)

        model_2 = model
        model_name_2 = self.get_option(preset, "bot_2", "model")
        if model_name_2:
            model_2 = window.core.models.get(model_name_2)
            bot_2_kwargs["model"] = model_2
        bot_2_name = self.get_option(preset, "bot_2", "name")
        bot_2_kwargs["bot_id"] = 2
        if experts:
            bot_2_kwargs["handoffs"] = experts
        bot_2 = self.get_agent(window, bot_2_kwargs)

        kwargs = {
            "max_turns": int(max_steps),
        }
        input_items: list[TResponseInputItem] = messages

        # reverse history if needed
        if use_partial_ctx:
            input_items = self.reverse_history(input_items)

        if not stream:
            while True:
                # -------- bot 1 --------
                kwargs["input"] = input_items
                if bridge.stopped():
                    bridge.on_stop(ctx)
                    break

                kwargs = self.prepare_model(model, window, previous_response_id, kwargs)
                ctx.set_agent_name(bot_1.name)
                result = await Runner.run(
                    bot_1,
                    **kwargs
                )
                response_id = result.last_response_id
                output_1 = final_output
                if verbose:
                    print("Final response:", result)

                final_output, last_response_id = window.core.api.openai.responses.unpack_agent_response(result, ctx)

                if bridge.stopped():
                    bridge.on_stop(ctx)
                    break

                # get and reverse items
                input_items = result.to_input_list()
                input_items = self.reverse_items(input_items, verbose=reverse_verbose)

                if use_partial_ctx:
                    ctx = bridge.on_next_ctx(
                        ctx=ctx,
                        input="",  # new ctx: input
                        output=output_1,  # prev ctx: output
                        response_id=response_id,
                        stream=False,
                    )

                # -------- bot 2 --------
                kwargs["input"] = input_items
                kwargs = self.prepare_model(model_2, window, previous_response_id, kwargs)
                ctx.set_agent_name(bot_2.name)
                result = await Runner.run(
                    bot_2,
                    **kwargs
                )
                response_id = result.last_response_id
                output_2 = final_output
                if verbose:
                    print("Final response:", result)

                final_output, last_response_id = window.core.api.openai.responses.unpack_agent_response(result, ctx)
                if bridge.stopped():
                    bridge.on_stop(ctx)
                    break

                # get and reverse items
                input_items = result.to_input_list()
                input_items = self.reverse_items(input_items, verbose=reverse_verbose)

                if use_partial_ctx:
                    ctx = bridge.on_next_ctx(
                        ctx=ctx,
                        input="", # new ctx: input
                        output=output_2,  # prev ctx: output
                        response_id=response_id,
                        stream=False,
                    )
        else:
            handler = StreamHandler(window, bridge)
            begin = True
            while True:
                # -------- bot 1 --------
                kwargs["input"] = input_items
                kwargs = self.prepare_model(model, window, previous_response_id, kwargs)
                ctx.set_agent_name(bot_1.name)
                result = Runner.run_streamed(
                    bot_1,
                    **kwargs
                )

                handler.reset()

                # bot 1 title
                # title = f"\n\n**{bot_1_name}**\n\n"
                # ctx.stream = title
                bridge.on_step(ctx, begin)
                begin = False
                handler.begin = begin
                # if not use_partial_ctx:
                    # handler.to_buffer(title)
                async for event in result.stream_events():
                    if bridge.stopped():
                        result.cancel()
                        bridge.on_stop(ctx)
                        break
                    final_output, response_id = handler.handle(event, ctx)

                output_1 = final_output

                if bridge.stopped():
                    bridge.on_stop(ctx)
                    break

                # get and reverse items
                input_items = result.to_input_list()
                input_items = self.reverse_items(input_items, verbose=reverse_verbose)

                if use_partial_ctx:
                    ctx = bridge.on_next_ctx(
                        ctx=ctx,
                        input="",  # new ctx: input
                        output=output_1,  # prev ctx: output
                        response_id=response_id,
                        stream=True,
                    )
                    handler.new()
                else:
                    bridge.on_next(ctx)

                # -------- bot 2 --------
                kwargs["input"] = input_items
                kwargs = self.prepare_model(model_2, window, previous_response_id, kwargs)
                ctx.set_agent_name(bot_2.name)
                result = Runner.run_streamed(
                    bot_2,
                    **kwargs
                )
                handler.reset()

                # bot 2 title
                # title = f"\n\n**{bot_2_name}**\n\n"
                # ctx.stream = title
                bridge.on_step(ctx, False)
                # if not use_partial_ctx:
                    # handler.to_buffer(title)
                async for event in result.stream_events():
                    if bridge.stopped():
                        result.cancel()
                        bridge.on_stop(ctx)
                        break
                    final_output, response_id = handler.handle(event, ctx)

                output_2 = final_output

                if bridge.stopped():
                    bridge.on_stop(ctx)
                    break

                # get and reverse items
                input_items = result.to_input_list()
                input_items = self.reverse_items(input_items, verbose=reverse_verbose)

                if use_partial_ctx:
                    ctx = bridge.on_next_ctx(
                        ctx=ctx,
                        input="", # new ctx: input
                        output=output_2,  # prev ctx: output
                        response_id=response_id,
                        stream=True,
                    )
                    handler.new()
                else:
                    bridge.on_next(ctx)

        return ctx, final_output, response_id


    def get_options(self) -> Dict[str, Any]:
        """
        Return Agent options

        :return: dict of options
        """
        return {
            "bot_1": {
                "label": "Bot 1",
                "options": {
                    "name": {
                        "type": "text",
                        "label": trans("agent.option.name"),
                        "default": "Bot 1",
                    },
                    "prompt": {
                        "type": "textarea",
                        "label": trans("agent.option.prompt"),
                        "description": trans("agent.option.prompt.b1.desc"),
                        "default": self.PROMPT_BOT_1,
                    },
                    "allow_local_tools": {
                        "type": "bool",
                        "label": trans("agent.option.tools.local"),
                        "description": trans("agent.option.tools.local.desc"),
                        "default": True,
                    },
                    "allow_remote_tools": {
                        "type": "bool",
                        "label": trans("agent.option.tools.remote"),
                        "description": trans("agent.option.tools.remote.desc"),
                        "default": True,
                    },
                }
            },
            "bot_2": {
                "label": "Bot 2",
                "options": {
                    "name": {
                        "type": "text",
                        "label": trans("agent.option.name"),
                        "default": "Bot 2",
                    },
                    "model": {
                        "label": trans("agent.option.model"),
                        "type": "combo",
                        "use": "models",
                        "default": "gpt-4o",
                    },
                    "prompt": {
                        "type": "textarea",
                        "label": trans("agent.option.prompt"),
                        "description": trans("agent.option.prompt.b2.desc"),
                        "default": self.PROMPT_BOT_2,
                    },
                    "allow_local_tools": {
                        "type": "bool",
                        "label": trans("agent.option.tools.local"),
                        "description": trans("agent.option.tools.local.desc"),
                        "default": True,
                    },
                    "allow_remote_tools": {
                        "type": "bool",
                        "label": trans("agent.option.tools.remote"),
                        "description": trans("agent.option.tools.remote.desc"),
                        "default": True,
                    },
                }
            },
        }





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.26 12:00:00                  #
# ================================================== #

from __future__ import annotations
from typing import Dict, List, Optional, Any
from dataclasses import dataclass, field


# ------------------------ Types registry (templates) ------------------------

@dataclass
class PropertySpec:
    id: str
    type: str
    name: Optional[str] = None
    editable: bool = True
    value: Any = None
    allowed_inputs: int = 0
    allowed_outputs: int = 0
    options: Optional[List[str]] = None
    placeholder: Optional[str] = None      # hint text for text editors
    description: Optional[str] = None      # tooltip/help text shown in UI


@dataclass
class NodeTypeSpec:
    type_name: str
    title: Optional[str] = None
    # UI-only human-readable label used for menus; never persisted nor used as an identifier
    display_name: Optional[str] = None
    properties: List[PropertySpec] = field(default_factory=list)
    # Below are optional extensions for agent-flow needs:
    base_id: Optional[str] = None        # base prefix for friendly ids, e.g. "agent"
    export_kind: Optional[str] = None    # short kind for export, e.g. "agent", "start"
    bg_color: Optional[str] = None       # optional per-type background color (CSS/hex)
    max_num: Optional[int] = None        # optional per-layout cap; None or <=0 means unlimited

class NodeTypeRegistry:
    """Registry for node type specifications. Extend/override in subclasses."""
    def __init__(self, empty: bool = False):
        self._types: Dict[str, NodeTypeSpec] = {}
        if not empty:
            self._install_default_types()

    def register(self, spec: NodeTypeSpec):
        self._types[spec.type_name] = spec

    def types(self) -> List[str]:
        return list(self._types.keys())

    def get(self, type_name: str) -> Optional[NodeTypeSpec]:
        return self._types.get(type_name)

    def display_name(self, type_name: str) -> str:
        """Return UI label for a type: spec.display_name if non-empty, otherwise type_name."""
        spec = self.get(type_name)
        if spec:
            dn = getattr(spec, "display_name", None)
            if isinstance(dn, str) and dn.strip():
                return dn
        return type_name

    def _install_default_types(self):
        # Example/basic nodes kept intact
        self.register(NodeTypeSpec(
            type_name="Value/Float",
            title="Float",
            properties=[
                PropertySpec(id="value", type="float", name="Value", editable=True, value=0.0,
                             allowed_inputs=0, allowed_outputs=1),
            ]
        ))
        self.register(NodeTypeSpec(
            type_name="Math/Add",
            title="Add",
            properties=[
                PropertySpec(id="A", type="float", name="A", editable=True, value=0.0, allowed_inputs=1, allowed_outputs=0),
                PropertySpec(id="B", type="float", name="B", editable=True, value=0.0, allowed_inputs=1, allowed_outputs=0),
                PropertySpec(id="result", type="float", name="Result", editable=False, value=None, allowed_inputs=0, allowed_outputs=1),
            ]
        ))
        # Tip: to allow multiple connections to an input or output, set allowed_inputs/allowed_outputs to -1.


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.27 20:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        # Endpoints / HTTP
        plugin.add_option(
            "api_base",
            type="text",
            value="https://openapi.tuyaeu.com",
            label="API base",
            description="Tuya API base (eu/us/cn/in: https://openapi.tuyaeu.com / tuyaus / tuyacn / tuyain).",
        )
        plugin.add_option(
            "http_timeout",
            type="int",
            value=30,
            label="HTTP timeout (s)",
            description="Requests timeout in seconds.",
        )
        plugin.add_option(
            "lang",
            type="text",
            value="en",
            label="Language",
            description="Language header (e.g., en, pl).",
        )

        # Credentials
        plugin.add_option(
            "tuya_client_id",
            type="text",
            value="",
            label="Tuya Client ID",
            description="Client ID from Tuya IoT Platform Cloud project.",
            secret=True,
        )
        plugin.add_option(
            "tuya_client_secret",
            type="text",
            value="",
            label="Tuya Client Secret",
            description="Client secret from Tuya IoT Platform Cloud project.",
            secret=True,
        )
        plugin.add_option(
            "tuya_uid",
            type="text",
            value="",
            label="Tuya UID (App Account)",
            description="UID of linked Tuya App account (required for listing devices).",
        )

        # Tokens (auto)
        plugin.add_option(
            "tuya_access_token",
            type="textarea",
            value="",
            label="(auto) Access token",
            description="Stored Tuya access token.",
            secret=True,
        )
        plugin.add_option(
            "tuya_refresh_token",
            type="textarea",
            value="",
            label="(auto) Refresh token",
            description="Stored Tuya refresh token (not always used).",
            secret=True,
        )
        plugin.add_option(
            "tuya_token_expires_in",
            type="text",
            value="",
            label="(auto) Expires in (s)",
            description="Token lifetime seconds.",
        )
        plugin.add_option(
            "tuya_token_expire_at",
            type="text",
            value="0",
            label="(auto) Expire at (epoch s)",
            description="Timestamp when token is considered expired.",
        )

        # Cache
        plugin.add_option(
            "tuya_cached_devices",
            type="textarea",
            value="[]",
            label="(auto) Cached devices",
            description="Cached devices for quick search.",
        )

        # ---------------- Commands ----------------

        # Auth
        plugin.add_cmd(
            "tuya_set_keys",
            instruction="Set Tuya Cloud credentials.",
            params=[
                {"name": "client_id", "type": "str", "required": True, "description": "Tuya Client ID"},
                {"name": "client_secret", "type": "str", "required": True, "description": "Tuya Client Secret"},
            ],
            enabled=True,
            description="Auth: set keys",
            tab="auth",
        )
        plugin.add_cmd(
            "tuya_set_uid",
            instruction="Set Tuya App Account UID (for listing devices).",
            params=[
                {"name": "uid", "type": "str", "required": True, "description": "UID of linked Tuya account"},
            ],
            enabled=True,
            description="Auth: set UID",
            tab="auth",
        )
        plugin.add_cmd(
            "tuya_token_get",
            instruction="Obtain access token (grant_type=1).",
            params=[],
            enabled=True,
            description="Auth: get token",
            tab="auth",
        )

        # Devices
        plugin.add_cmd(
            "tuya_devices_list",
            instruction="List devices for UID.",
            params=[
                {"name": "uid", "type": "str", "required": False, "description": "Override UID (optional)"},
                {"name": "page_no", "type": "int", "required": False, "description": "Page number (default 1)"},
                {"name": "page_size", "type": "int", "required": False, "description": "Page size (default 100)"},
            ],
            enabled=True,
            description="Devices: list",
            tab="devices",
        )
        plugin.add_cmd(
            "tuya_device_get",
            instruction="Get device info.",
            params=[{"name": "device_id", "type": "str", "required": True, "description": "Device ID"}],
            enabled=True,
            description="Devices: get info",
            tab="devices",
        )
        plugin.add_cmd(
            "tuya_device_status",
            instruction="Get device status (DP values).",
            params=[{"name": "device_id", "type": "str", "required": True, "description": "Device ID"}],
            enabled=True,
            description="Devices: status",
            tab="devices",
        )
        plugin.add_cmd(
            "tuya_device_functions",
            instruction="Get device supported functions (DP codes).",
            params=[{"name": "device_id", "type": "str", "required": True, "description": "Device ID"}],
            enabled=True,
            description="Devices: functions",
            tab="devices",
        )
        plugin.add_cmd(
            "tuya_find_device",
            instruction="Find device(s) by name (uses cached device list).",
            params=[{"name": "name", "type": "str", "required": True, "description": "Substring match"}],
            enabled=True,
            description="Devices: find by name",
            tab="devices",
        )

        # Control
        plugin.add_cmd(
            "tuya_device_set",
            instruction="Set a device DP value or multiple values.",
            params=[
                {"name": "device_id", "type": "str", "required": True, "description": "Device ID"},
                {"name": "code", "type": "str", "required": False, "description": "DP code (if single)"},
                {"name": "value", "type": "str", "required": False, "description": "Value for 'code'"},
                {"name": "codes", "type": "dict", "required": False, "description": "Dict of {code: value}"},
            ],
            enabled=True,
            description="Control: set DP(s)",
            tab="control",
        )
        plugin.add_cmd(
            "tuya_device_send",
            instruction="Send raw commands list to device.",
            params=[
                {"name": "device_id", "type": "str", "required": True, "description": "Device ID"},
                {"name": "commands", "type": "list", "required": True, "description": "[{'code':'','value':..},...]"},
            ],
            enabled=True,
            description="Control: send commands",
            tab="control",
        )
        plugin.add_cmd(
            "tuya_device_on",
            instruction="Turn device ON (guesses switch code if not provided).",
            params=[
                {"name": "device_id", "type": "str", "required": True, "description": "Device ID"},
                {"name": "code", "type": "str", "required": False, "description": "Preferred switch code"},
            ],
            enabled=True,
            description="Control: on",
            tab="control",
        )
        plugin.add_cmd(
            "tuya_device_off",
            instruction="Turn device OFF (guesses switch code if not provided).",
            params=[
                {"name": "device_id", "type": "str", "required": True, "description": "Device ID"},
                {"name": "code", "type": "str", "required": False, "description": "Preferred switch code"},
            ],
            enabled=True,
            description="Control: off",
            tab="control",
        )
        plugin.add_cmd(
            "tuya_device_toggle",
            instruction="Toggle device state (reads current boolean DP).",
            params=[
                {"name": "device_id", "type": "str", "required": True, "description": "Device ID"},
                {"name": "code", "type": "str", "required": False, "description": "Preferred switch code"},
            ],
            enabled=True,
            description="Control: toggle",
            tab="control",
        )

        # Sensors
        plugin.add_cmd(
            "tuya_sensors_read",
            instruction="Read and normalize common sensor values.",
            params=[{"name": "device_id", "type": "str", "required": True, "description": "Sensor device ID"}],
            enabled=True,
            description="Sensors: read",
            tab="sensors",
        )